@@ -35,7 +35,6 @@ void spice_channel_handle_set_ack(SpiceChannel *channel, SpiceMsgIn *in)
     c->message_ack_window = c->message_ack_count = ack->window;
     c->marshallers->msgc_ack_sync(out->marshaller, &sync);
     spice_msg_out_send_internal(out);
-    spice_msg_out_unref(out);
 }
 
 /* coroutine context */
@@ -48,7 +47,6 @@ void spice_channel_handle_ping(SpiceChannel *channel, SpiceMsgIn *in)
 
     c->marshallers->msgc_pong(pong->marshaller, ping);
     spice_msg_out_send_internal(pong);
-    spice_msg_out_unref(pong);
 }
 
 /* coroutine context */
@@ -92,14 +90,51 @@ void spice_channel_handle_disconnect(SpiceChannel *channel, SpiceMsgIn *in)
                 disconnect->time_stamp, disconnect->reason);
 }
 
+typedef struct WaitForChannelData
+{
+    SpiceWaitForChannel *wait;
+    SpiceChannel *channel;
+} WaitForChannelData;
+
+/* coroutine and main context */
+static gboolean wait_for_channel(gpointer data)
+{
+    WaitForChannelData *wfc = data;
+    SpiceChannelPrivate *c = wfc->channel->priv;
+    SpiceChannel *wait_channel;
+
+    wait_channel = spice_session_lookup_channel(c->session, wfc->wait->channel_id, wfc->wait->channel_type);
+    g_return_val_if_fail(wait_channel != NULL, TRUE);
+
+    if (wait_channel->priv->last_message_serial >= wfc->wait->message_serial)
+        return TRUE;
+
+    return FALSE;
+}
+
 /* coroutine context */
 G_GNUC_INTERNAL
 void spice_channel_handle_wait_for_channels(SpiceChannel *channel, SpiceMsgIn *in)
 {
-    /* SpiceChannelPrivate *c = channel->priv;
-       SpiceMsgWaitForChannels *wfc = spice_msg_in_parsed(in); */
+    SpiceChannelPrivate *c = channel->priv;
+    SpiceMsgWaitForChannels *wfc = spice_msg_in_parsed(in);
+    int i;
 
-    SPICE_DEBUG("%s TODO", __FUNCTION__);
+    g_return_if_fail(spice_header_get_msg_size(in->header, channel->priv->use_mini_header) >=
+                     sizeof(*wfc) + wfc->wait_count * sizeof(wfc->wait_list[0]));
+
+    for (i = 0; i < wfc->wait_count; ++i) {
+        WaitForChannelData data = {
+            .wait = wfc->wait_list + i,
+            .channel = channel
+        };
+
+        SPICE_DEBUG("waiting for serial %" PRIu64 " (%d/%d)", data.wait->message_serial, i + 1, wfc->wait_count);
+        if (g_coroutine_condition_wait(&c->coroutine, wait_for_channel, &data))
+            SPICE_DEBUG("waiting for serial %"  PRIu64 ", done", data.wait->message_serial);
+        else
+            SPICE_DEBUG("waiting for serial %" PRIu64 ", cancelled", data.wait->message_serial);
+    }
 }
 
 static void
@@ -130,15 +165,16 @@ void spice_channel_handle_migrate(SpiceChannel *channel, SpiceMsgIn *in)
 
         out = spice_msg_out_new(SPICE_CHANNEL(channel), SPICE_MSGC_MIGRATE_FLUSH_MARK);
         spice_msg_out_send_internal(out);
-        spice_msg_out_unref(out);
         SPICE_CHANNEL_GET_CLASS(channel)->iterate_write(channel);
     }
     if (mig->flags & SPICE_MIGRATE_NEED_DATA_TRANSFER) {
         spice_channel_recv_msg(channel, get_msg_handler, &data);
         if (!data) {
             g_warning("expected SPICE_MSG_MIGRATE_DATA, got empty message");
-        } else if (data->header.type != SPICE_MSG_MIGRATE_DATA) {
-            g_warning("expected SPICE_MSG_MIGRATE_DATA, got %d", data->header.type);
+        } else if (spice_header_get_msg_type(data->header, c->use_mini_header) !=
+                   SPICE_MSG_MIGRATE_DATA) {
+            g_warning("expected SPICE_MSG_MIGRATE_DATA, got %d",
+                      spice_header_get_msg_type(data->header, c->use_mini_header));
         }
     }
 
@@ -146,8 +182,8 @@ void spice_channel_handle_migrate(SpiceChannel *channel, SpiceMsgIn *in)
 
     if ((mig->flags & SPICE_MIGRATE_NEED_DATA_TRANSFER) && (data != NULL)) {
         out = spice_msg_out_new(SPICE_CHANNEL(channel), SPICE_MSGC_MIGRATE_DATA);
-        spice_marshaller_add(out->marshaller, data->data, data->header.size);
+        spice_marshaller_add(out->marshaller, data->data,
+                             spice_header_get_msg_size(data->header, c->use_mini_header));
         spice_msg_out_send_internal(out);
-        spice_msg_out_unref(out);
     }
 }
