@@ -69,7 +69,6 @@ void stream_mjpeg_data(display_stream *st)
     int height = info->stream_height;
     uint8_t *dest;
     uint8_t *lines[4];
-    int i, j;
 
     dest = malloc(width * height * 4);
 
@@ -95,10 +94,19 @@ void stream_mjpeg_data(display_stream *st)
 #endif
     // TODO: in theory should check cinfo.output_height match with our height
     jpeg_start_decompress(&st->mjpeg_cinfo);
-    for (i = 0; i < height; ) {
-        // according to header
-        g_return_if_fail(st->mjpeg_cinfo.rec_outbuf_height <= 4);
-        for (j = 0; j < st->mjpeg_cinfo.rec_outbuf_height; j++) {
+    /* rec_outbuf_height is the recommended size of the output buffer we
+     * pass to libjpeg for optimum performance
+     */
+    if (st->mjpeg_cinfo.rec_outbuf_height > G_N_ELEMENTS(lines)) {
+        jpeg_abort_decompress(&st->mjpeg_cinfo);
+        g_return_if_reached();
+    }
+
+    while (st->mjpeg_cinfo.output_scanline < st->mjpeg_cinfo.output_height) {
+        /* only used when JCS_EXTENSIONS is undefined */
+        G_GNUC_UNUSED unsigned int lines_read;
+
+        for (unsigned int j = 0; j < st->mjpeg_cinfo.rec_outbuf_height; j++) {
             lines[j] = dest;
 #ifdef JCS_EXTENSIONS
             dest += 4 * width;
@@ -106,16 +114,14 @@ void stream_mjpeg_data(display_stream *st)
             dest += 3 * width;
 #endif
         }
-        j = jpeg_read_scanlines(&st->mjpeg_cinfo, lines,
+        lines_read = jpeg_read_scanlines(&st->mjpeg_cinfo, lines,
                                 st->mjpeg_cinfo.rec_outbuf_height);
-        // this shouldn't happen either..
-        g_return_if_fail(j == st->mjpeg_cinfo.rec_outbuf_height);
 #ifndef JCS_EXTENSIONS
         {
             uint8_t *s = lines[0];
-            uint32_t *d = (uint32_t *)&st->out_frame[i * width * 4];
+            uint32_t *d = (uint32_t *)s;
 
-            for (j = j * width; j > 0; ) {
+            for (unsigned int j = lines_read * width; j > 0; ) {
                 j -= 1; // reverse order, bad for cache?
                 d[j] = s[j * 3 + 0] << 16 |
                     s[j * 3 + 1] << 8 |
@@ -123,8 +129,7 @@ void stream_mjpeg_data(display_stream *st)
             }
         }
 #endif
-        i += st->mjpeg_cinfo.rec_outbuf_height;
-        dest = &st->out_frame[i * width * 4];
+        dest = &st->out_frame[st->mjpeg_cinfo.output_scanline * width * 4];
     }
     jpeg_finish_decompress(&st->mjpeg_cinfo);
 }
@@ -133,4 +138,6 @@ G_GNUC_INTERNAL
 void stream_mjpeg_cleanup(display_stream *st)
 {
     jpeg_destroy_decompress(&st->mjpeg_cinfo);
+    free(st->out_frame);
+    st->out_frame = NULL;
 }
