@@ -73,6 +73,7 @@ struct _SpiceDisplayChannelPrivate {
 #ifdef WIN32
     HDC dc;
 #endif
+    gboolean                    migrate_wait_primary;
 };
 
 G_DEFINE_TYPE(SpiceDisplayChannel, spice_display_channel, SPICE_TYPE_CHANNEL)
@@ -103,6 +104,8 @@ static void clear_surfaces(SpiceChannel *channel);
 static void clear_streams(SpiceChannel *channel);
 static display_surface *find_surface(SpiceDisplayChannelPrivate *c, int surface_id);
 static gboolean display_stream_render(display_stream *st);
+static void spice_display_channel_reset(SpiceChannel *channel, gboolean migrating);
+static void destroy_canvas(display_surface *surface);
 
 /* ------------------------------------------------------------------ */
 
@@ -119,13 +122,17 @@ static void spice_display_channel_dispose(GObject *object)
         G_OBJECT_CLASS(spice_display_channel_parent_class)->dispose(object);
 }
 
-static void spice_display_channel_finalize(GObject *obj)
+static void spice_display_channel_finalize(GObject *object)
 {
-    clear_surfaces(SPICE_CHANNEL(obj));
-    clear_streams(SPICE_CHANNEL(obj));
+    SpiceDisplayChannelPrivate *c = SPICE_DISPLAY_CHANNEL(object)->priv;
+
+    c->migrate_wait_primary = FALSE;
+
+    clear_surfaces(SPICE_CHANNEL(object));
+    clear_streams(SPICE_CHANNEL(object));
 
     if (G_OBJECT_CLASS(spice_display_channel_parent_class)->finalize)
-        G_OBJECT_CLASS(spice_display_channel_parent_class)->finalize(obj);
+        G_OBJECT_CLASS(spice_display_channel_parent_class)->finalize(object);
 }
 
 static void spice_display_channel_constructed(GObject *object)
@@ -181,6 +188,20 @@ static void spice_display_set_property(GObject      *object,
     }
 }
 
+/* main or coroutine context */
+static void spice_display_channel_reset(SpiceChannel *channel, gboolean migrating)
+{
+    SpiceDisplayChannelPrivate *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
+
+    /* palettes, images, and glz_window are cleared in the session */
+
+    c->migrate_wait_primary = migrating;
+    clear_streams(channel);
+    clear_surfaces(channel);
+
+    SPICE_CHANNEL_CLASS(spice_display_channel_parent_class)->channel_reset(channel, migrating);
+}
+
 static void spice_display_channel_class_init(SpiceDisplayChannelClass *klass)
 {
     GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
@@ -194,6 +215,7 @@ static void spice_display_channel_class_init(SpiceDisplayChannelClass *klass)
 
     channel_class->handle_msg   = spice_display_handle_msg;
     channel_class->channel_up   = spice_display_channel_up;
+    channel_class->channel_reset = spice_display_channel_reset;
 
     g_object_class_install_property
         (gobject_class, PROP_HEIGHT,
@@ -382,18 +404,44 @@ static void image_put(SpiceImageCache *cache, uint64_t id, pixman_image_t *image
     item->ptr = pixman_image_ref(image);
 }
 
+typedef struct _WaitImageData
+{
+    gboolean lossy;
+    SpiceImageCache *cache;
+    uint64_t id;
+    pixman_image_t *image;
+} WaitImageData;
+
+static gboolean wait_image(gpointer data)
+{
+    display_cache_item *item;
+    WaitImageData *wait = data;
+    SpiceDisplayChannelPrivate *c =
+        SPICE_CONTAINEROF(wait->cache, SpiceDisplayChannelPrivate, image_cache);
+
+    item = cache_find(c->images, wait->id);
+    if (item == NULL ||
+        (item->lossy && !wait->lossy))
+        return FALSE;
+
+    cache_used(c->images, item);
+    wait->image = pixman_image_ref(item->ptr);
+
+    return TRUE;
+}
+
 static pixman_image_t *image_get(SpiceImageCache *cache, uint64_t id)
 {
-    SpiceDisplayChannelPrivate *c =
-        SPICE_CONTAINEROF(cache, SpiceDisplayChannelPrivate, image_cache);
-    display_cache_item *item;
+    WaitImageData wait = {
+        .lossy = TRUE,
+        .cache = cache,
+        .id = id,
+        .image = NULL
+    };
+    if (!g_coroutine_condition_wait(g_coroutine_self(), wait_image, &wait))
+        SPICE_DEBUG("wait image got cancelled");
 
-    item = cache_find(c->images, id);
-    if (item) {
-        cache_used(c->images, item);
-        return pixman_image_ref(item->ptr);
-    }
-    return NULL;
+    return wait.image;
 }
 
 static void image_remove(SpiceImageCache *cache, uint64_t id)
@@ -489,20 +537,16 @@ static void image_replace_lossy(SpiceImageCache *cache, uint64_t id,
 
 static pixman_image_t* image_get_lossless(SpiceImageCache *cache, uint64_t id)
 {
-    SpiceDisplayChannelPrivate *c =
-        SPICE_CONTAINEROF(cache, SpiceDisplayChannelPrivate, image_cache);
-    display_cache_item *item;
+    WaitImageData wait = {
+        .lossy = FALSE,
+        .cache = cache,
+        .id = id,
+        .image = NULL
+    };
+    if (!g_coroutine_condition_wait(g_coroutine_self(), wait_image, &wait))
+        SPICE_DEBUG("wait lossless got cancelled");
 
-    item = cache_find(c->images, id);
-    if (!item)
-        return NULL;
-
-    /* TODO: shared_cache.hpp does wait until it is lossless..., is
-       that necessary? */
-    g_warn_if_fail(item->lossy == FALSE);
-
-    cache_used(c->images, item);
-    return pixman_image_ref(item->ptr);
+    return wait.image;
 }
 #endif
 
@@ -555,7 +599,6 @@ static void spice_display_channel_init(SpiceDisplayChannel *channel)
     SpiceDisplayChannelPrivate *c;
 
     c = channel->priv = SPICE_DISPLAY_CHANNEL_GET_PRIVATE(channel);
-    memset(c, 0, sizeof(*c));
 
     ring_init(&c->surfaces);
     c->image_cache.ops = &image_cache_ops;
@@ -573,7 +616,26 @@ static int create_canvas(SpiceChannel *channel, display_surface *surface)
     SpiceDisplayChannelPrivate *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
 
     if (surface->primary) {
+        display_surface *primary = find_surface(c, 0);
+
+        if (primary) {
+            if (c->migrate_wait_primary &&
+                primary->width == surface->width &&
+                primary->height == surface->height) {
+                c->migrate_wait_primary = FALSE;
+                SPICE_DEBUG("Reusing existing primary surface");
+                return 0;
+            }
+
+            emit_main_context(channel, SPICE_DISPLAY_PRIMARY_DESTROY);
+            ring_remove(&primary->link);
+            destroy_canvas(primary);
+            free(primary);
+        }
+
         SPICE_DEBUG("display: create primary canvas");
+        c->migrate_wait_primary = FALSE;
+
 #ifdef HAVE_SYS_SHM_H
         surface->shmid = shmget(IPC_PRIVATE, surface->size, IPC_CREAT | 0777);
         if (surface->shmid >= 0) {
@@ -620,6 +682,12 @@ static int create_canvas(SpiceChannel *channel, display_surface *surface)
                                              surface->zlib_decoder);
 
     g_return_val_if_fail(surface->canvas != NULL, 0);
+    ring_add(&c->surfaces, &surface->link);
+
+    if (surface->primary)
+        emit_main_context(channel, SPICE_DISPLAY_PRIMARY_CREATE,
+                          surface->format, surface->width, surface->height,
+                          surface->stride, surface->shmid, surface->data);
     return 0;
 }
 
@@ -668,9 +736,15 @@ static void clear_surfaces(SpiceChannel *channel)
     display_surface *surface;
     RingItem *item;
 
-    while (!ring_is_empty(&c->surfaces)) {
-        item = ring_get_head(&c->surfaces);
+    for (item = ring_get_head(&c->surfaces); item != NULL; ) {
         surface = SPICE_CONTAINEROF(item, display_surface, link);
+        item = ring_next(&c->surfaces, item);
+
+        if (c->migrate_wait_primary && surface->primary) {
+            SPICE_DEBUG("Try to keep exisiting primary surface during migration");
+            continue;
+        }
+
         ring_remove(&surface->link);
         destroy_canvas(surface);
         free(surface);
@@ -692,17 +766,24 @@ static void emit_invalidate(SpiceChannel *channel, SpiceRect *bbox)
 static void spice_display_channel_up(SpiceChannel *channel)
 {
     SpiceMsgOut *out;
-    SpiceMsgcDisplayInit init = {
-        .pixmap_cache_id            = 1,
-        .pixmap_cache_size          = DISPLAY_PIXMAP_CACHE,
-        .glz_dictionary_id          = 1,
-        .glz_dictionary_window_size = GLZ_WINDOW_SIZE,
-    };
+    SpiceSession *s = spice_channel_get_session(channel);
+    SpiceMsgcDisplayInit init;
+    int cache_size;
+    int glz_window_size;
 
+    g_object_get(s,
+                 "cache-size", &cache_size,
+                 "glz-window-size", &glz_window_size,
+                 NULL);
+    SPICE_DEBUG("%s: cache_size %d, glz_window_size %d (bytes)", __FUNCTION__,
+                cache_size, glz_window_size);
+    init.pixmap_cache_id = 1;
+    init.glz_dictionary_id = 1;
+    init.pixmap_cache_size = cache_size / 4; /* pixels */
+    init.glz_dictionary_window_size = glz_window_size / 4; /* pixels */
     out = spice_msg_out_new(channel, SPICE_MSGC_DISPLAY_INIT);
     out->marshallers->msgc_display_init(out->marshaller, &init);
     spice_msg_out_send_internal(out);
-    spice_msg_out_unref(out);
 }
 
 #define DRAW(type) {                                                    \
@@ -722,17 +803,10 @@ static void display_handle_mode(SpiceChannel *channel, SpiceMsgIn *in)
 {
     SpiceDisplayChannelPrivate *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
     SpiceMsgDisplayMode *mode = spice_msg_in_parsed(in);
-    display_surface *surface = find_surface(c, 0);
+    display_surface *surface;
 
     g_warn_if_fail(c->mark == FALSE);
 
-    if (surface) {
-        emit_main_context(channel, SPICE_DISPLAY_PRIMARY_DESTROY);
-        ring_remove(&surface->link);
-        destroy_canvas(surface);
-        free(surface);
-    }
-
     surface = spice_new0(display_surface, 1);
     surface->format  = mode->bits == 32 ?
         SPICE_SURFACE_FMT_32_xRGB : SPICE_SURFACE_FMT_16_555;
@@ -742,15 +816,11 @@ static void display_handle_mode(SpiceChannel *channel, SpiceMsgIn *in)
     surface->size    = surface->height * surface->stride;
     surface->primary = true;
     create_canvas(channel, surface);
-    emit_main_context(channel, SPICE_DISPLAY_PRIMARY_CREATE,
-                      surface->format, surface->width, surface->height,
-                      surface->stride, surface->shmid, surface->data);
 #ifdef HAVE_SYS_SHM_H
     if (surface->shmid != -1) {
         shmctl(surface->shmid, IPC_RMID, 0);
     }
 #endif
-    ring_add(&c->surfaces, &surface->link);
 }
 
 /* coroutine context */
@@ -823,6 +893,8 @@ static void display_handle_inv_list(SpiceChannel *channel, SpiceMsgIn *in)
 /* coroutine context */
 static void display_handle_inv_pixmap_all(SpiceChannel *channel, SpiceMsgIn *in)
 {
+    spice_channel_handle_wait_for_channels(channel, in);
+
     spice_session_images_clear(spice_channel_get_session(channel));
 }
 
@@ -1203,9 +1275,6 @@ static void display_handle_surface_create(SpiceChannel *channel, SpiceMsgIn *in)
     if (create->flags == SPICE_SURFACE_FLAGS_PRIMARY) {
         surface->primary = true;
         create_canvas(channel, surface);
-        emit_main_context(channel, SPICE_DISPLAY_PRIMARY_CREATE,
-                          surface->format, surface->width, surface->height,
-                          surface->stride, surface->shmid, surface->data);
         if (c->mark_false_event_id != 0) {
             g_source_remove(c->mark_false_event_id);
             c->mark_false_event_id = FALSE;
@@ -1214,8 +1283,6 @@ static void display_handle_surface_create(SpiceChannel *channel, SpiceMsgIn *in)
         surface->primary = false;
         create_canvas(channel, surface);
     }
-
-    ring_add(&c->surfaces, &surface->link);
 }
 
 static gboolean display_mark_false(gpointer data)
