@@ -15,6 +15,7 @@
    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, see <http://www.gnu.org/licenses/>.
 */
+#include "glib-compat.h"
 #include "spice-client.h"
 #include "spice-common.h"
 #include "spice-marshal.h"
@@ -74,6 +75,7 @@ struct _SpiceMainChannelPrivate  {
     GQueue                      *agent_msg_queue;
 
     guint                       switch_host_delayed_id;
+    guint                       migrate_delayed_id;
 };
 
 typedef struct spice_migrate spice_migrate;
@@ -121,6 +123,7 @@ static guint signals[SPICE_MAIN_LAST_SIGNAL];
 
 static void spice_main_handle_msg(SpiceChannel *channel, SpiceMsgIn *msg);
 static void agent_send_msg_queue(SpiceMainChannel *channel);
+static void agent_free_msg_queue(SpiceMainChannel *channel);
 static void migrate_channel_event_cb(SpiceChannel *channel, SpiceChannelEvent event,
                                      gpointer data);
 
@@ -155,8 +158,9 @@ static void spice_main_channel_init(SpiceMainChannel *channel)
     SpiceMainChannelPrivate *c;
 
     c = channel->priv = SPICE_MAIN_CHANNEL_GET_PRIVATE(channel);
-    memset(c, 0, sizeof(*c));
     c->agent_msg_queue = g_queue_new();
+
+    spice_channel_set_capability(SPICE_CHANNEL(channel), SPICE_MAIN_CAP_SEMI_SEAMLESS_MIGRATE);
 }
 
 static void spice_main_get_property(GObject    *object,
@@ -241,6 +245,11 @@ static void spice_main_channel_dispose(GObject *obj)
         c->switch_host_delayed_id = 0;
     }
 
+    if (c->migrate_delayed_id) {
+        g_source_remove(c->migrate_delayed_id);
+        c->migrate_delayed_id = 0;
+    }
+
     if (G_OBJECT_CLASS(spice_main_channel_parent_class)->dispose)
         G_OBJECT_CLASS(spice_main_channel_parent_class)->dispose(obj);
 }
@@ -250,7 +259,7 @@ static void spice_main_channel_finalize(GObject *obj)
     SpiceMainChannelPrivate *c = SPICE_MAIN_CHANNEL(obj)->priv;
 
     g_free(c->agent_msg_data);
-    g_queue_free(c->agent_msg_queue);
+    agent_free_msg_queue(SPICE_MAIN_CHANNEL(obj));
 
     if (G_OBJECT_CLASS(spice_main_channel_parent_class)->finalize)
         G_OBJECT_CLASS(spice_main_channel_parent_class)->finalize(obj);
@@ -265,6 +274,26 @@ static void spice_channel_iterate_write(SpiceChannel *channel)
         SPICE_CHANNEL_CLASS(spice_main_channel_parent_class)->iterate_write(channel);
 }
 
+/* main or coroutine context */
+static void spice_main_channel_reset(SpiceChannel *channel, gboolean migrating)
+{
+    SpiceMainChannelPrivate *c = SPICE_MAIN_CHANNEL(channel)->priv;
+
+    c->agent_connected = FALSE;
+    c->agent_caps_received = FALSE;
+    c->agent_display_config_sent = FALSE;
+    c->agent_tokens = 0;
+    c->agent_msg_pos = 0;
+    g_free(c->agent_msg_data);
+    c->agent_msg_data = NULL;
+    c->agent_msg_size = 0;
+
+    agent_free_msg_queue(SPICE_MAIN_CHANNEL(channel));
+    c->agent_msg_queue = g_queue_new();
+
+    SPICE_CHANNEL_CLASS(spice_main_channel_parent_class)->channel_reset(channel, migrating);
+}
+
 static void spice_main_channel_class_init(SpiceMainChannelClass *klass)
 {
     GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
@@ -277,6 +306,7 @@ static void spice_main_channel_class_init(SpiceMainChannelClass *klass)
 
     channel_class->handle_msg    = spice_main_handle_msg;
     channel_class->iterate_write = spice_channel_iterate_write;
+    channel_class->channel_reset = spice_main_channel_reset;
 
     /**
      * SpiceMainChannel:mouse-mode:
@@ -419,7 +449,7 @@ static void spice_main_channel_class_init(SpiceMainChannelClass *klass)
     signals[SPICE_MAIN_CLIPBOARD] =
         g_signal_new("main-clipboard",
                      G_OBJECT_CLASS_TYPE(gobject_class),
-                     G_SIGNAL_RUN_LAST,
+                     G_SIGNAL_RUN_LAST | G_SIGNAL_DEPRECATED,
                      0,
                      NULL, NULL,
                      g_cclosure_user_marshal_VOID__UINT_POINTER_UINT,
@@ -458,7 +488,7 @@ static void spice_main_channel_class_init(SpiceMainChannelClass *klass)
     signals[SPICE_MAIN_CLIPBOARD_GRAB] =
         g_signal_new("main-clipboard-grab",
                      G_OBJECT_CLASS_TYPE(gobject_class),
-                     G_SIGNAL_RUN_LAST,
+                     G_SIGNAL_RUN_LAST | G_SIGNAL_DEPRECATED,
                      0,
                      NULL, NULL,
                      g_cclosure_user_marshal_BOOLEAN__POINTER_UINT,
@@ -501,7 +531,7 @@ static void spice_main_channel_class_init(SpiceMainChannelClass *klass)
     signals[SPICE_MAIN_CLIPBOARD_REQUEST] =
         g_signal_new("main-clipboard-request",
                      G_OBJECT_CLASS_TYPE(gobject_class),
-                     G_SIGNAL_RUN_LAST,
+                     G_SIGNAL_RUN_LAST | G_SIGNAL_DEPRECATED,
                      0,
                      NULL, NULL,
                      g_cclosure_user_marshal_BOOLEAN__UINT,
@@ -542,7 +572,7 @@ static void spice_main_channel_class_init(SpiceMainChannelClass *klass)
     signals[SPICE_MAIN_CLIPBOARD_RELEASE] =
         g_signal_new("main-clipboard-release",
                      G_OBJECT_CLASS_TYPE(gobject_class),
-                     G_SIGNAL_RUN_LAST,
+                     G_SIGNAL_RUN_LAST | G_SIGNAL_DEPRECATED,
                      0,
                      NULL, NULL,
                      g_cclosure_marshal_VOID__VOID,
@@ -706,6 +736,23 @@ static void do_emit_main_context(GObject *object, int signum, gpointer params)
 
 /* ------------------------------------------------------------------ */
 
+static void agent_free_msg_queue(SpiceMainChannel *channel)
+{
+    SpiceMainChannelPrivate *c = channel->priv;
+    SpiceMsgOut *out;
+
+    if (!c->agent_msg_queue)
+        return;
+
+    while (!g_queue_is_empty(c->agent_msg_queue)) {
+        out = g_queue_pop_head(c->agent_msg_queue);
+        spice_msg_out_unref(out);
+    }
+
+    g_queue_free(c->agent_msg_queue);
+    c->agent_msg_queue = NULL;
+}
+
 /* coroutine context */
 static void agent_send_msg_queue(SpiceMainChannel *channel)
 {
@@ -717,7 +764,6 @@ static void agent_send_msg_queue(SpiceMainChannel *channel)
         c->agent_tokens--;
         out = g_queue_pop_head(c->agent_msg_queue);
         spice_msg_out_send_internal(out);
-        spice_msg_out_unref(out);
     }
 }
 
@@ -1023,7 +1069,6 @@ static void agent_start(SpiceMainChannel *channel)
     out = spice_msg_out_new(SPICE_CHANNEL(channel), SPICE_MSGC_MAIN_AGENT_START);
     out->marshallers->msgc_main_agent_start(out->marshaller, &agent_start);
     spice_msg_out_send_internal(out);
-    spice_msg_out_unref(out);
 
     if (c->agent_connected) {
         agent_announce_caps(channel);
@@ -1054,15 +1099,17 @@ static void set_mouse_mode(SpiceMainChannel *channel, uint32_t supported, uint32
     }
 
     /* switch to client mode if possible */
-    if ((supported & SPICE_MOUSE_MODE_CLIENT) && (current != SPICE_MOUSE_MODE_CLIENT)) {
+    if (!spice_channel_get_read_only(SPICE_CHANNEL(channel)) &&
+        supported & SPICE_MOUSE_MODE_CLIENT &&
+        current != SPICE_MOUSE_MODE_CLIENT) {
         SpiceMsgcMainMouseModeRequest req = {
             .mode = SPICE_MOUSE_MODE_CLIENT,
         };
         SpiceMsgOut *out;
+
         out = spice_msg_out_new(SPICE_CHANNEL(channel), SPICE_MSGC_MAIN_MOUSE_MODE_REQUEST);
         out->marshallers->msgc_main_mouse_mode_request(out->marshaller, &req);
         spice_msg_out_send_internal(out);
-        spice_msg_out_unref(out);
     }
 }
 
@@ -1077,19 +1124,21 @@ static void main_handle_init(SpiceChannel *channel, SpiceMsgIn *in)
     session = spice_channel_get_session(channel);
     spice_session_set_connection_id(session, init->session_id);
 
-    out = spice_msg_out_new(SPICE_CHANNEL(channel), SPICE_MSGC_MAIN_ATTACH_CHANNELS);
-    spice_msg_out_send_internal(out);
-    spice_msg_out_unref(out);
-
     set_mouse_mode(SPICE_MAIN_CHANNEL(channel), init->supported_mouse_modes,
                    init->current_mouse_mode);
 
-    c->agent_tokens = init->agent_tokens;
-    if (init->agent_connected) {
-        agent_start(SPICE_MAIN_CHANNEL(channel));
-    }
-
     spice_session_set_mm_time(session, init->multi_media_time);
+    spice_session_set_caches_hints(session, init->ram_hint, init->display_channels_hint);
+
+    c->agent_tokens = init->agent_tokens;
+    if (init->agent_connected)
+        agent_start(SPICE_MAIN_CHANNEL(channel));
+
+    if (spice_session_migrate_after_main_init(session))
+        return;
+
+    out = spice_msg_out_new(SPICE_CHANNEL(channel), SPICE_MSGC_MAIN_ATTACH_CHANNELS);
+    spice_msg_out_send_internal(out);
 }
 
 /* coroutine context */
@@ -1351,13 +1400,15 @@ static void migrate_channel_new_cb(SpiceSession *s, SpiceChannel *channel, gpoin
                      G_CALLBACK(migrate_channel_event_cb), data);
 }
 
-static void migrate_channel_connect(spice_migrate *mig, int type, int id)
+static SpiceChannel* migrate_channel_connect(spice_migrate *mig, int type, int id)
 {
     SPICE_DEBUG("migrate_channel_connect %d:%d", type, id);
 
     SpiceChannel *newc = spice_channel_new(mig->session, type, id);
     spice_channel_connect(newc);
     mig->nchannels++;
+
+    return newc;
 }
 
 /* main context */
@@ -1432,7 +1483,7 @@ static gboolean migrate_connect(gpointer data)
     mig->session = spice_session_new_from_session(session);
 
     if ((c->peer_hdr.major_version == 1) &&
-        (c->peer_hdr.minor_version < 2)) {
+        (c->peer_hdr.minor_version < 1)) {
         OldRedMigrationBegin *info = (OldRedMigrationBegin *)mig->info;
         SPICE_DEBUG("migrate_begin old %s %d %d",
                     info->host, info->port, info->sport);
@@ -1440,19 +1491,34 @@ static gboolean migrate_connect(gpointer data)
         sport = info->sport;
         host = info->host;
     } else {
-        GByteArray *pubkey = g_byte_array_new();
         SpiceMsgMainMigrationBegin *info = mig->info;
         SPICE_DEBUG("migrate_begin %d %s %d %d",
                     info->host_size, info->host_data, info->port, info->sport);
         port = info->port;
         sport = info->sport;
         host = (char*)info->host_data;
-        g_byte_array_append(pubkey, info->pub_key_data, info->pub_key_size);
-        g_object_set(mig->session,
-                     "pubkey", pubkey,
-                     "verify", SPICE_SESSION_VERIFY_PUBKEY,
-                     NULL);
-        g_byte_array_unref(pubkey);
+
+        if ((c->peer_hdr.major_version == 1) ||
+            (c->peer_hdr.major_version == 2 && c->peer_hdr.minor_version < 1)) {
+            GByteArray *pubkey = g_byte_array_new();
+
+            g_byte_array_append(pubkey, info->pub_key_data, info->pub_key_size);
+            g_object_set(mig->session,
+                         "pubkey", pubkey,
+                         "verify", SPICE_SESSION_VERIFY_PUBKEY,
+                         NULL);
+            g_byte_array_unref(pubkey);
+        } else {
+            gchar *subject = g_alloca(info->cert_subject_size + 1);
+            strncpy(subject, (const char*)info->cert_subject_data, info->cert_subject_size);
+            subject[info->cert_subject_size] = '\0';
+
+            // session data are already copied
+            g_object_set(mig->session,
+                         "cert-subject", subject,
+                         "verify", SPICE_SESSION_VERIFY_SUBJECT,
+                         NULL);
+        }
     }
 
     if (g_getenv("SPICE_MIG_HOST"))
@@ -1469,7 +1535,7 @@ static gboolean migrate_connect(gpointer data)
 
     /* the migration process is in 2 steps, first the main channel and
        then the rest of the channels */
-    migrate_channel_connect(mig, SPICE_CHANNEL_MAIN, 0);
+    mig->session->priv->cmain = migrate_channel_connect(mig, SPICE_CHANNEL_MAIN, 0);
 
     return FALSE;
 }
@@ -1505,7 +1571,33 @@ static void main_handle_migrate_begin(SpiceChannel *channel, SpiceMsgIn *in)
 
     out = spice_msg_out_new(SPICE_CHANNEL(channel), reply_type);
     spice_msg_out_send(out);
-    spice_msg_out_unref(out);
+}
+
+/* main context */
+static gboolean migrate_delayed(gpointer data)
+{
+    SpiceChannel *channel = data;
+    SpiceMainChannelPrivate *c = SPICE_MAIN_CHANNEL(channel)->priv;
+
+    g_warn_if_fail(c->migrate_delayed_id != 0);
+    c->migrate_delayed_id = 0;
+
+    spice_session_migrate_end(channel->priv->session);
+
+    return FALSE;
+}
+
+/* coroutine context */
+static void main_handle_migrate_end(SpiceChannel *channel, SpiceMsgIn *in)
+{
+    SpiceMainChannelPrivate *c = SPICE_MAIN_CHANNEL(channel)->priv;
+
+    SPICE_DEBUG("migrate end");
+
+    g_return_if_fail(c->migrate_delayed_id == 0);
+    g_return_if_fail(spice_channel_test_capability(channel, SPICE_MAIN_CAP_SEMI_SEAMLESS_MIGRATE));
+
+    c->migrate_delayed_id = g_idle_add(migrate_delayed, channel);
 }
 
 /* main context */
@@ -1586,6 +1678,7 @@ static const spice_msg_handler main_handlers[] = {
     [ SPICE_MSG_MAIN_AGENT_TOKEN ]         = main_handle_agent_token,
 
     [ SPICE_MSG_MAIN_MIGRATE_BEGIN ]       = main_handle_migrate_begin,
+    [ SPICE_MSG_MAIN_MIGRATE_END ]         = main_handle_migrate_end,
     [ SPICE_MSG_MAIN_MIGRATE_CANCEL ]      = main_handle_migrate_cancel,
     [ SPICE_MSG_MAIN_MIGRATE_SWITCH_HOST ] = main_handle_migrate_switch_host,
 };
@@ -1617,7 +1710,7 @@ static gboolean timer_set_display(gpointer data)
     c->timer_id = 0;
     if (c->agent_connected)
         spice_main_send_monitor_config(SPICE_MAIN_CHANNEL(channel));
-    spice_channel_wakeup(channel);
+    spice_channel_wakeup(channel, FALSE);
 
     return false;
 }
@@ -1667,6 +1760,7 @@ void spice_main_set_display(SpiceMainChannel *channel, int id,
  *
  * Deprecated: 0.6: use spice_main_clipboard_selection_grab() instead.
  **/
+G_GNUC_DEPRECATED_FOR(spice_main_clipboard_selection_grab)
 void spice_main_clipboard_grab(SpiceMainChannel *channel, guint32 *types, int ntypes)
 {
     spice_main_clipboard_selection_grab(channel, VD_AGENT_CLIPBOARD_SELECTION_CLIPBOARD, types, ntypes);
@@ -1690,7 +1784,7 @@ void spice_main_clipboard_selection_grab(SpiceMainChannel *channel, guint select
     g_return_if_fail(SPICE_IS_MAIN_CHANNEL(channel));
 
     agent_clipboard_grab(channel, selection, types, ntypes);
-    spice_channel_wakeup(SPICE_CHANNEL(channel));
+    spice_channel_wakeup(SPICE_CHANNEL(channel), FALSE);
 }
 
 /**
@@ -1702,6 +1796,7 @@ void spice_main_clipboard_selection_grab(SpiceMainChannel *channel, guint select
  *
  * Deprecated: 0.6: use spice_main_clipboard_selection_release() instead.
  **/
+G_GNUC_DEPRECATED_FOR(spice_main_clipboard_selection_release)
 void spice_main_clipboard_release(SpiceMainChannel *channel)
 {
     spice_main_clipboard_selection_release(channel, VD_AGENT_CLIPBOARD_SELECTION_CLIPBOARD);
@@ -1728,7 +1823,7 @@ void spice_main_clipboard_selection_release(SpiceMainChannel *channel, guint sel
         return;
 
     agent_clipboard_release(channel, selection);
-    spice_channel_wakeup(SPICE_CHANNEL(channel));
+    spice_channel_wakeup(SPICE_CHANNEL(channel), FALSE);
 }
 
 /**
@@ -1742,6 +1837,7 @@ void spice_main_clipboard_selection_release(SpiceMainChannel *channel, guint sel
  *
  * Deprecated: 0.6: use spice_main_clipboard_selection_notify() instead.
  **/
+G_GNUC_DEPRECATED_FOR(spice_main_clipboard_selection_notify)
 void spice_main_clipboard_notify(SpiceMainChannel *channel,
                                  guint32 type, const guchar *data, size_t size)
 {
@@ -1768,7 +1864,7 @@ void spice_main_clipboard_selection_notify(SpiceMainChannel *channel, guint sele
     g_return_if_fail(SPICE_IS_MAIN_CHANNEL(channel));
 
     agent_clipboard_notify(channel, selection, type, data, size);
-    spice_channel_wakeup(SPICE_CHANNEL(channel));
+    spice_channel_wakeup(SPICE_CHANNEL(channel), FALSE);
 }
 
 /**
@@ -1781,6 +1877,7 @@ void spice_main_clipboard_selection_notify(SpiceMainChannel *channel, guint sele
  *
  * Deprecated: 0.6: use spice_main_clipboard_selection_request() instead.
  **/
+G_GNUC_DEPRECATED_FOR(spice_main_clipboard_selection_request)
 void spice_main_clipboard_request(SpiceMainChannel *channel, guint32 type)
 {
     spice_main_clipboard_selection_request(channel, VD_AGENT_CLIPBOARD_SELECTION_CLIPBOARD, type);
@@ -1803,7 +1900,7 @@ void spice_main_clipboard_selection_request(SpiceMainChannel *channel, guint sel
     g_return_if_fail(SPICE_IS_MAIN_CHANNEL(channel));
 
     agent_clipboard_request(channel, selection, type);
-    spice_channel_wakeup(SPICE_CHANNEL(channel));
+    spice_channel_wakeup(SPICE_CHANNEL(channel), FALSE);
 }
 
 /**
