@@ -91,10 +91,8 @@ static void channel_up(SpiceChannel *channel);
 
 static void spice_record_channel_init(SpiceRecordChannel *channel)
 {
-    SpiceRecordChannelPrivate *c;
+    channel->priv = SPICE_RECORD_CHANNEL_GET_PRIVATE(channel);
 
-    c = channel->priv = SPICE_RECORD_CHANNEL_GET_PRIVATE(channel);
-    memset(c, 0, sizeof(*c));
     spice_channel_set_capability(SPICE_CHANNEL(channel), SPICE_RECORD_CAP_CELT_0_5_1);
     spice_channel_set_capability(SPICE_CHANNEL(channel), SPICE_RECORD_CAP_VOLUME);
 }
@@ -165,6 +163,26 @@ static void spice_record_channel_set_property(GObject      *gobject,
     }
 }
 
+static void channel_reset(SpiceChannel *channel, gboolean migrating)
+{
+    SpiceRecordChannelPrivate *c = SPICE_RECORD_CHANNEL(channel)->priv;
+
+    g_free(c->last_frame);
+    c->last_frame = NULL;
+
+    if (c->celt_encoder) {
+        celt051_encoder_destroy(c->celt_encoder);
+        c->celt_encoder = NULL;
+    }
+
+    if (c->celt_mode) {
+        celt051_mode_destroy(c->celt_mode);
+        c->celt_mode = NULL;
+    }
+
+    SPICE_CHANNEL_CLASS(spice_record_channel_parent_class)->channel_reset(channel, migrating);
+}
+
 static void spice_record_channel_class_init(SpiceRecordChannelClass *klass)
 {
     GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
@@ -175,6 +193,7 @@ static void spice_record_channel_class_init(SpiceRecordChannelClass *klass)
     gobject_class->set_property = spice_record_channel_set_property;
     channel_class->handle_msg   = spice_record_handle_msg;
     channel_class->channel_up   = channel_up;
+    channel_class->channel_reset = channel_reset;
 
     g_object_class_install_property
         (gobject_class, PROP_NCHANNELS,
@@ -279,6 +298,8 @@ static void spice_record_mode(SpiceRecordChannel *channel, uint32_t time,
     SpiceMsgOut *msg;
 
     g_return_if_fail(channel != NULL);
+    if (spice_channel_get_read_only(SPICE_CHANNEL(channel)))
+        return;
 
     m.mode = mode;
     m.time = time;
@@ -288,7 +309,6 @@ static void spice_record_mode(SpiceRecordChannel *channel, uint32_t time,
     msg = spice_msg_out_new(SPICE_CHANNEL(channel), SPICE_MSGC_RECORD_MODE);
     msg->marshallers->msgc_record_mode(msg->marshaller, &m);
     spice_msg_out_send(msg);
-    spice_msg_out_unref(msg);
 }
 
 /* coroutine context */
@@ -311,13 +331,14 @@ static void spice_record_start_mark(SpiceRecordChannel *channel, uint32_t time)
     SpiceMsgOut *msg;
 
     g_return_if_fail(channel != NULL);
+    if (spice_channel_get_read_only(SPICE_CHANNEL(channel)))
+        return;
 
     m.time = time;
 
     msg = spice_msg_out_new(SPICE_CHANNEL(channel), SPICE_MSGC_RECORD_START_MARK);
     msg->marshallers->msgc_record_start_mark(msg->marshaller, &m);
     spice_msg_out_send(msg);
-    spice_msg_out_unref(msg);
 }
 
 /**
@@ -338,6 +359,7 @@ void spice_record_send_data(SpiceRecordChannel *channel, gpointer data,
     uint8_t *celt_buf = NULL;
 
     g_return_if_fail(channel != NULL);
+    g_return_if_fail(spice_channel_get_read_only(SPICE_CHANNEL(channel)) == FALSE);
 
     rc = channel->priv;
 
@@ -395,7 +417,6 @@ void spice_record_send_data(SpiceRecordChannel *channel, gpointer data,
         msg->marshallers->msgc_record_data(msg->marshaller, &p);
         spice_marshaller_add(msg->marshaller, frame, frame_size);
         spice_msg_out_send(msg);
-        spice_msg_out_unref(msg);
 
         if (rc->last_frame_current == rc->frame_bytes)
             rc->last_frame_current = 0;
