@@ -144,31 +144,56 @@ static void spice_smartcard_channel_init(SpiceSmartcardChannel *channel)
 
 static void spice_smartcard_channel_finalize(GObject *obj)
 {
-    SpiceSmartcardChannel *channel = SPICE_SMARTCARD_CHANNEL(obj);
+    SpiceSmartcardChannelPrivate *c = SPICE_SMARTCARD_CHANNEL_GET_PRIVATE(obj);
 
-    if (channel->priv->pending_card_insertions != NULL) {
-        g_hash_table_destroy(channel->priv->pending_card_insertions);
-        channel->priv->pending_card_insertions = NULL;
+    if (c->pending_card_insertions != NULL) {
+        g_hash_table_destroy(c->pending_card_insertions);
+        c->pending_card_insertions = NULL;
     }
-    if (channel->priv->pending_reader_removals != NULL) {
-        g_hash_table_destroy(channel->priv->pending_reader_removals);
-        channel->priv->pending_reader_removals = NULL;
+    if (c->pending_reader_removals != NULL) {
+        g_hash_table_destroy(c->pending_reader_removals);
+        c->pending_reader_removals = NULL;
     }
-    if (channel->priv->message_queue != NULL) {
-        g_queue_foreach(channel->priv->message_queue,
-                        (GFunc)smartcard_message_free, NULL);
-        g_queue_free(channel->priv->message_queue);
-        channel->priv->message_queue = NULL;
+    if (c->message_queue != NULL) {
+        g_queue_foreach(c->message_queue, (GFunc)smartcard_message_free, NULL);
+        g_queue_free(c->message_queue);
+        c->message_queue = NULL;
     }
-    if (channel->priv->in_flight_message != NULL) {
-        smartcard_message_free(channel->priv->in_flight_message);
-        channel->priv->in_flight_message = NULL;
+    if (c->in_flight_message != NULL) {
+        smartcard_message_free(c->in_flight_message);
+        c->in_flight_message = NULL;
     }
 
+    g_list_free(c->pending_reader_additions);
+    c->pending_reader_additions = NULL;
+
     if (G_OBJECT_CLASS(spice_smartcard_channel_parent_class)->finalize)
         G_OBJECT_CLASS(spice_smartcard_channel_parent_class)->finalize(obj);
 }
 
+static void spice_smartcard_channel_reset(SpiceChannel *channel, gboolean migrating)
+{
+    SpiceSmartcardChannelPrivate *c = SPICE_SMARTCARD_CHANNEL_GET_PRIVATE(channel);
+
+    g_hash_table_remove_all(c->pending_card_insertions);
+    g_hash_table_remove_all(c->pending_reader_removals);
+
+    if (c->message_queue != NULL) {
+        g_queue_foreach(c->message_queue, (GFunc)smartcard_message_free, NULL);
+        g_queue_clear(c->message_queue);
+    }
+
+    if (c->in_flight_message != NULL) {
+        smartcard_message_free(c->in_flight_message);
+        c->in_flight_message = NULL;
+    }
+
+    g_list_free(c->pending_reader_additions);
+    c->pending_reader_additions = NULL;
+
+    SPICE_CHANNEL_CLASS(spice_smartcard_channel_parent_class)->channel_reset(channel, migrating);
+}
+
 static void spice_smartcard_channel_class_init(SpiceSmartcardChannelClass *klass)
 {
     GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
@@ -177,6 +202,7 @@ static void spice_smartcard_channel_class_init(SpiceSmartcardChannelClass *klass
     gobject_class->finalize     = spice_smartcard_channel_finalize;
     channel_class->handle_msg   = spice_smartcard_handle_msg;
     channel_class->channel_up   = spice_smartcard_channel_up;
+    channel_class->channel_reset = spice_smartcard_channel_reset;
 
     g_type_class_add_private(klass, sizeof(SpiceSmartcardChannelPrivate));
 }
@@ -191,7 +217,8 @@ static const spice_msg_handler smartcard_handlers[] = {
 static void
 smartcard_message_free(SpiceSmartcardChannelMessage *message)
 {
-    spice_msg_out_unref(message->message);
+    if (message->message)
+        spice_msg_out_unref(message->message);
     g_slice_free(SpiceSmartcardChannelMessage, message);
 }
 
@@ -264,15 +291,14 @@ smartcard_message_new(VSCMsgType msg_type, SpiceMsgOut *msg_out)
 static void
 smartcard_message_complete_in_flight(SpiceSmartcardChannel *channel)
 {
-    if (channel->priv->in_flight_message == NULL) {
-        g_return_if_fail(g_queue_is_empty(channel->priv->message_queue));
-        return;
-    }
+    g_return_if_fail(channel->priv->in_flight_message != NULL);
 
     smartcard_message_free(channel->priv->in_flight_message);
     channel->priv->in_flight_message = g_queue_pop_head(channel->priv->message_queue);
-    if (channel->priv->in_flight_message != NULL)
+    if (channel->priv->in_flight_message != NULL) {
         spice_msg_out_send(channel->priv->in_flight_message->message);
+        channel->priv->in_flight_message->message = NULL;
+    }
 }
 
 static void smartcard_message_send(SpiceSmartcardChannel *channel,
@@ -281,11 +307,13 @@ static void smartcard_message_send(SpiceSmartcardChannel *channel,
 {
     SpiceSmartcardChannelMessage *message;
 
+    if (spice_channel_get_read_only(SPICE_CHANNEL(channel)))
+        return;
+
     SPICE_DEBUG("smartcard: send message %d, %s",
                 msg_type, queue ? "queued" : "now");
     if (!queue) {
         spice_msg_out_send(msg_out);
-        spice_msg_out_unref(msg_out);
         return;
     }
 
@@ -293,7 +321,8 @@ static void smartcard_message_send(SpiceSmartcardChannel *channel,
     if (channel->priv->in_flight_message == NULL) {
         g_return_if_fail(g_queue_is_empty(channel->priv->message_queue));
         channel->priv->in_flight_message = message;
-        spice_msg_out_send(msg_out);
+        spice_msg_out_send(channel->priv->in_flight_message->message);
+        channel->priv->in_flight_message->message = NULL;
     } else {
         g_queue_push_tail(channel->priv->message_queue, message);
     }
@@ -427,7 +456,7 @@ static void spice_smartcard_channel_up_cb(GObject *source_object,
     spice_smartcard_manager_init_finish(SPICE_SESSION(source_object),
                                         res, &error);
     if (error)
-        g_warning(error->message);
+        g_warning("%s", error->message);
     g_clear_error(&error);
 }
 
