@@ -40,18 +40,21 @@
 
 G_BEGIN_DECLS
 
+#define MAX_SPICE_DATA_HEADER_SIZE sizeof(SpiceDataHeader)
+
 struct _SpiceMsgOut {
     int                   refcount;
     SpiceChannel          *channel;
     SpiceMessageMarshallers *marshallers;
     SpiceMarshaller       *marshaller;
-    SpiceDataHeader       *header;
+    uint8_t               *header;
+    gboolean              ro_check;
 };
 
 struct _SpiceMsgIn {
     int                   refcount;
     SpiceChannel          *channel;
-    SpiceDataHeader       header;
+    uint8_t               header[MAX_SPICE_DATA_HEADER_SIZE];
     uint8_t               *data;
     int                   hpos,dpos;
     uint8_t               *parsed;
@@ -85,18 +88,21 @@ struct _SpiceChannelPrivate {
     unsigned int                sasl_decoded_offset;
 #endif
 
+    gboolean                    use_mini_header;
+    uint64_t                    out_serial;
+    uint64_t                    in_serial;
+
     /* not swapped */
     SpiceSession                *session;
-    struct coroutine            coroutine;
+    GCoroutine                  coroutine;
     int                         fd;
     gboolean                    has_error;
     guint                       connect_delayed_id;
 
-    int                         wait_interruptible;
-    struct wait_queue           wait;
-    guint8                      *xmit_buffer;
-    int                         xmit_buffer_capacity;
-    int                         xmit_buffer_size;
+    GQueue                      xmit_queue;
+    gboolean                    xmit_queue_blocked;
+    GStaticMutex                xmit_queue_lock;
+    guint                       xmit_queue_wakeup_id;
 
     char                        name[16];
     enum spice_channel_state    state;
@@ -108,7 +114,6 @@ struct _SpiceChannelPrivate {
     int                         connection_id;
     int                         channel_id;
     int                         channel_type;
-    int                         serial;
     SpiceLinkHeader             link_hdr;
     SpiceLinkMess               link_msg;
     SpiceLinkHeader             peer_hdr;
@@ -125,6 +130,7 @@ struct _SpiceChannelPrivate {
     GArray                      *remote_common_caps;
 
     gsize                       total_read_bytes;
+    uint64_t                    last_message_serial;
 };
 
 SpiceMsgIn *spice_msg_in_new(SpiceChannel *channel);
@@ -144,8 +150,11 @@ void spice_msg_out_send(SpiceMsgOut *out);
 void spice_msg_out_send_internal(SpiceMsgOut *out);
 void spice_msg_out_hexdump(SpiceMsgOut *out, unsigned char *data, int len);
 
+uint16_t spice_header_get_msg_type(uint8_t *header, gboolean is_mini_header);
+uint32_t spice_header_get_msg_size(uint8_t *header, gboolean is_mini_header);
+
 void spice_channel_up(SpiceChannel *channel);
-void spice_channel_wakeup(SpiceChannel *channel);
+void spice_channel_wakeup(SpiceChannel *channel, gboolean cancel);
 
 SpiceSession* spice_channel_get_session(SpiceChannel *channel);
 
@@ -166,6 +175,9 @@ gint spice_channel_get_channel_id(SpiceChannel *channel);
 gint spice_channel_get_channel_type(SpiceChannel *channel);
 void spice_channel_swap(SpiceChannel *channel, SpiceChannel *swap);
 void spice_channel_set_common_capability(SpiceChannel *channel, guint32 cap);
+gboolean spice_channel_get_read_only(SpiceChannel *channel);
+
+void spice_channel_reset(SpiceChannel *channel, gboolean migrating);
 
 /* coroutine context */
 #define emit_main_context(object, event, args...)                       \
