@@ -33,6 +33,7 @@
 #endif
 #ifdef HAVE_NETINET_IN_H
 #include <netinet/in.h>
+#include <netinet/tcp.h> // TCP_NODELAY
 #endif
 #ifdef HAVE_ARPA_INET_H
 #include <arpa/inet.h>
@@ -42,9 +43,10 @@
 #include "gio-coroutine.h"
 
 static void spice_channel_handle_msg(SpiceChannel *channel, SpiceMsgIn *msg);
-static void spice_channel_send_msg(SpiceChannel *channel, SpiceMsgOut *out, gboolean buffered);
+static void spice_channel_write_msg(SpiceChannel *channel, SpiceMsgOut *out);
 static void spice_channel_send_link(SpiceChannel *channel);
 static void channel_disconnect(SpiceChannel *channel);
+static void channel_reset(SpiceChannel *channel, gboolean migrating);
 
 /**
  * SECTION:spice-channel
@@ -96,7 +98,8 @@ static void spice_channel_init(SpiceChannel *channel)
 
     c = channel->priv = SPICE_CHANNEL_GET_PRIVATE(channel);
 
-    c->serial = 1;
+    c->out_serial = 1;
+    c->in_serial = 1;
     c->fd = -1;
     strcpy(c->name, "?");
     c->caps = g_array_new(FALSE, TRUE, sizeof(guint32));
@@ -104,6 +107,9 @@ static void spice_channel_init(SpiceChannel *channel)
     c->remote_caps = g_array_new(FALSE, TRUE, sizeof(guint32));
     c->remote_common_caps = g_array_new(FALSE, TRUE, sizeof(guint32));
     spice_channel_set_common_capability(channel, SPICE_COMMON_CAP_PROTOCOL_AUTH_SELECTION);
+    spice_channel_set_common_capability(channel, SPICE_COMMON_CAP_MINI_HEADER);
+    g_queue_init(&c->xmit_queue);
+    g_static_mutex_init(&c->xmit_queue_lock);
 }
 
 static void spice_channel_constructed(GObject *gobject)
@@ -248,6 +254,7 @@ static void spice_channel_class_init(SpiceChannelClass *klass)
     klass->iterate_write = spice_channel_iterate_write;
     klass->iterate_read  = spice_channel_iterate_read;
     klass->channel_disconnect = channel_disconnect;
+    klass->channel_reset = channel_reset;
 
     gobject_class->constructed  = spice_channel_constructed;
     gobject_class->dispose      = spice_channel_dispose;
@@ -309,10 +316,10 @@ static void spice_channel_class_init(SpiceChannelClass *klass)
                      G_SIGNAL_RUN_FIRST,
                      G_STRUCT_OFFSET(SpiceChannelClass, channel_event),
                      NULL, NULL,
-                     g_cclosure_marshal_VOID__INT,
+                     g_cclosure_marshal_VOID__ENUM,
                      G_TYPE_NONE,
                      1,
-                     G_TYPE_INT);
+                     SPICE_TYPE_CHANNEL_EVENT);
 
     /**
      * SpiceChannel::open-fd:
@@ -340,6 +347,101 @@ static void spice_channel_class_init(SpiceChannelClass *klass)
     SSL_load_error_strings();
 }
 
+/* ---------------------------------------------------------------- */
+/* private header api                                               */
+
+static inline void spice_header_set_msg_type(uint8_t *header, gboolean is_mini_header,
+                                             uint16_t type)
+{
+    if (is_mini_header) {
+        ((SpiceMiniDataHeader *)header)->type = type;
+    } else {
+        ((SpiceDataHeader *)header)->type = type;
+    }
+}
+
+static inline void spice_header_set_msg_size(uint8_t *header, gboolean is_mini_header,
+                                             uint32_t size)
+{
+    if (is_mini_header) {
+        ((SpiceMiniDataHeader *)header)->size = size;
+    } else {
+        ((SpiceDataHeader *)header)->size = size;
+    }
+}
+
+G_GNUC_INTERNAL
+uint16_t spice_header_get_msg_type(uint8_t *header, gboolean is_mini_header)
+{
+    if (is_mini_header) {
+        return ((SpiceMiniDataHeader *)header)->type;
+    } else {
+        return ((SpiceDataHeader *)header)->type;
+    }
+}
+
+G_GNUC_INTERNAL
+uint32_t spice_header_get_msg_size(uint8_t *header, gboolean is_mini_header)
+{
+    if (is_mini_header) {
+        return ((SpiceMiniDataHeader *)header)->size;
+    } else {
+        return ((SpiceDataHeader *)header)->size;
+    }
+}
+
+static inline int spice_header_get_header_size(gboolean is_mini_header)
+{
+    return is_mini_header ? sizeof(SpiceMiniDataHeader) : sizeof(SpiceDataHeader);
+}
+
+static inline void spice_header_set_msg_serial(uint8_t *header, gboolean is_mini_header,
+                                               uint64_t serial)
+{
+    if (!is_mini_header) {
+        ((SpiceDataHeader *)header)->serial = serial;
+    }
+}
+
+static inline void spice_header_reset_msg_sub_list(uint8_t *header, gboolean is_mini_header)
+{
+    if (!is_mini_header) {
+        ((SpiceDataHeader *)header)->sub_list = 0;
+    }
+}
+
+static inline uint64_t spice_header_get_in_msg_serial(SpiceMsgIn *in)
+{
+    SpiceChannelPrivate *c = in->channel->priv;
+    uint8_t *header = in->header;
+
+    if (c->use_mini_header) {
+        return c->in_serial;
+    } else {
+        return ((SpiceDataHeader *)header)->serial;
+    }
+}
+
+static inline uint64_t spice_header_get_out_msg_serial(SpiceMsgOut *out)
+{
+    SpiceChannelPrivate *c = out->channel->priv;
+
+    if (c->use_mini_header) {
+        return c->out_serial;
+    } else {
+        return ((SpiceDataHeader *)out->header)->serial;
+    }
+}
+
+static inline uint32_t spice_header_get_msg_sub_list(uint8_t *header, gboolean is_mini_header)
+{
+    if (is_mini_header) {
+        return 0;
+    } else {
+        return ((SpiceDataHeader *)header)->sub_list;
+    }
+}
+
 /* ---------------------------------------------------------------- */
 /* private msg api                                                  */
 
@@ -353,6 +455,7 @@ SpiceMsgIn *spice_msg_in_new(SpiceChannel *channel)
     in = spice_new0(SpiceMsgIn, 1);
     in->refcount = 1;
     in->channel  = channel;
+
     return in;
 }
 
@@ -365,8 +468,8 @@ SpiceMsgIn *spice_msg_in_sub_new(SpiceChannel *channel, SpiceMsgIn *parent,
     g_return_val_if_fail(channel != NULL, NULL);
 
     in = spice_msg_in_new(channel);
-    in->header.type = sub->type;
-    in->header.size = sub->size;
+    spice_header_set_msg_type(in->header, channel->priv->use_mini_header, sub->type);
+    spice_header_set_msg_size(in->header, channel->priv->use_mini_header, sub->size);
     in->data = (uint8_t*)(sub+1);
     in->dpos = sub->size;
     in->parent = parent;
@@ -405,7 +508,7 @@ int spice_msg_in_type(SpiceMsgIn *in)
 {
     g_return_val_if_fail(in != NULL, -1);
 
-    return in->header.type;
+    return spice_header_get_msg_type(in->header, in->channel->priv->use_mini_header);
 }
 
 G_GNUC_INTERNAL
@@ -449,8 +552,10 @@ void spice_msg_in_hexdump(SpiceMsgIn *in)
     SpiceChannelPrivate *c = in->channel->priv;
 
     fprintf(stderr, "--\n<< hdr: %s serial %" PRIu64 " type %d size %d sub-list %d\n",
-            c->name, in->header.serial, in->header.type,
-            in->header.size, in->header.sub_list);
+            c->name, spice_header_get_in_msg_serial(in),
+            spice_header_get_msg_type(in->header, c->use_mini_header),
+            spice_header_get_msg_size(in->header, c->use_mini_header),
+            spice_header_get_msg_sub_list(in->header, c->use_mini_header));
     hexdump("<< msg", in->data, in->dpos);
 }
 
@@ -460,11 +565,38 @@ void spice_msg_out_hexdump(SpiceMsgOut *out, unsigned char *data, int len)
     SpiceChannelPrivate *c = out->channel->priv;
 
     fprintf(stderr, "--\n>> hdr: %s serial %" PRIu64 " type %d size %d sub-list %d\n",
-            c->name, out->header->serial, out->header->type,
-            out->header->size, out->header->sub_list);
+            c->name,
+            spice_header_get_out_msg_serial(out),
+            spice_header_get_msg_type(out->header, c->use_mini_header),
+            spice_header_get_msg_size(out->header, c->use_mini_header),
+            spice_header_get_msg_sub_list(out->header, c->use_mini_header));
     hexdump(">> msg", data, len);
 }
 
+static gboolean msg_check_read_only (int channel_type, int msg_type)
+{
+    if (msg_type < 100) // those are the common messages
+        return FALSE;
+
+    switch (channel_type) {
+    /* messages allowed to be sent in read-only mode */
+    case SPICE_CHANNEL_MAIN:
+        switch (msg_type) {
+        case SPICE_MSGC_MAIN_CLIENT_INFO:
+        case SPICE_MSGC_MAIN_MIGRATE_CONNECTED:
+        case SPICE_MSGC_MAIN_MIGRATE_CONNECT_ERROR:
+        case SPICE_MSGC_MAIN_ATTACH_CHANNELS:
+        case SPICE_MSGC_MAIN_MIGRATE_END:
+            return FALSE;
+        }
+        break;
+    case SPICE_CHANNEL_DISPLAY:
+        return FALSE;
+    }
+
+    return TRUE;
+}
+
 G_GNUC_INTERNAL
 SpiceMsgOut *spice_msg_out_new(SpiceChannel *channel, int type)
 {
@@ -476,15 +608,19 @@ SpiceMsgOut *spice_msg_out_new(SpiceChannel *channel, int type)
     out = spice_new0(SpiceMsgOut, 1);
     out->refcount = 1;
     out->channel  = channel;
+    out->ro_check = msg_check_read_only(c->channel_type, type);
 
     out->marshallers = c->marshallers;
     out->marshaller = spice_marshaller_new();
-    out->header = (SpiceDataHeader *)
-        spice_marshaller_reserve_space(out->marshaller, sizeof(SpiceDataHeader));
-    spice_marshaller_set_base(out->marshaller, sizeof(SpiceDataHeader));
-    out->header->serial = c->serial++;
-    out->header->type = type;
-    out->header->sub_list = 0;
+
+    out->header = spice_marshaller_reserve_space(out->marshaller,
+                                                 spice_header_get_header_size(c->use_mini_header));
+    spice_marshaller_set_base(out->marshaller, spice_header_get_header_size(c->use_mini_header));
+    spice_header_set_msg_type(out->header, c->use_mini_header, type);
+    spice_header_set_msg_serial(out->header, c->use_mini_header, c->out_serial);
+    spice_header_reset_msg_sub_list(out->header, c->use_mini_header);
+
+    c->out_serial++;
     return out;
 }
 
@@ -509,17 +645,59 @@ void spice_msg_out_unref(SpiceMsgOut *out)
 }
 
 /* system context */
+static gboolean spice_channel_idle_wakeup(gpointer user_data)
+{
+    SpiceChannel *channel = SPICE_CHANNEL(user_data);
+    SpiceChannelPrivate *c = channel->priv;
+
+    /*
+     * Note:
+     *
+     * - This must be done before the wakeup as that may eventually
+     *   call channel_reset() which checks this.
+     * - The lock calls are really necessary, this fixes the following race:
+     *   1) usb-event-thread calls spice_msg_out_send()
+     *   2) spice_msg_out_send calls g_timeout_add_full(...)
+     *   3) we run, set xmit_queue_wakeup_id to 0
+     *   4) spice_msg_out_send stores the result of g_timeout_add_full() in
+     *      xmit_queue_wakeup_id, overwriting the 0 we just stored
+     *   5) xmit_queue_wakeup_id now says there is a wakeup pending which is
+     *      false
+     */
+    g_static_mutex_lock(&c->xmit_queue_lock);
+    c->xmit_queue_wakeup_id = 0;
+    g_static_mutex_unlock(&c->xmit_queue_lock);
+
+    spice_channel_wakeup(channel, FALSE);
+
+    return FALSE;
+}
+
+/* any context (system/co-routine/usb-event-thread) */
 G_GNUC_INTERNAL
 void spice_msg_out_send(SpiceMsgOut *out)
 {
     g_return_if_fail(out != NULL);
+    g_return_if_fail(out->channel != NULL);
 
-    out->header->size =
-        spice_marshaller_get_total_size(out->marshaller) - sizeof(SpiceDataHeader);
-    spice_channel_send_msg(out->channel, out, TRUE);
+    g_static_mutex_lock(&out->channel->priv->xmit_queue_lock);
+    if (!out->channel->priv->xmit_queue_blocked) {
+        gboolean was_empty;
 
-    /* TODO: we currently flush/wakeup immediately all buffered messages */
-    spice_channel_wakeup(out->channel);
+        was_empty = g_queue_is_empty(&out->channel->priv->xmit_queue);
+        g_queue_push_tail(&out->channel->priv->xmit_queue, out);
+
+        /* One wakeup is enough to empty the entire queue -> only do a wakeup
+           if the queue was empty, and there isn't one pending already. */
+        if (was_empty && !out->channel->priv->xmit_queue_wakeup_id) {
+            out->channel->priv->xmit_queue_wakeup_id =
+                /* Use g_timeout_add_full so that can specify the priority */
+                g_timeout_add_full(G_PRIORITY_HIGH, 0,
+                                   spice_channel_idle_wakeup,
+                                   out->channel, NULL);
+        }
+    }
+    g_static_mutex_unlock(&out->channel->priv->xmit_queue_lock);
 }
 
 /* coroutine context */
@@ -528,9 +706,7 @@ void spice_msg_out_send_internal(SpiceMsgOut *out)
 {
     g_return_if_fail(out != NULL);
 
-    out->header->size =
-        spice_marshaller_get_total_size(out->marshaller) - sizeof(SpiceDataHeader);
-    spice_channel_send_msg(out->channel, out, FALSE);
+    spice_channel_write_msg(out->channel, out);
 }
 
 /* ---------------------------------------------------------------- */
@@ -602,7 +778,8 @@ static void spice_channel_flush_wire(SpiceChannel *channel,
         }
         if (ret == -1) {
             if (cond != 0) {
-                g_io_wait(c->sock, cond);
+                g_coroutine_socket_wait(&c->coroutine, c->sock, cond);
+                continue;
             } else {
                 SPICE_DEBUG("Closing the channel: spice_channel_flush %d", errno);
                 c->has_error = TRUE;
@@ -656,6 +833,37 @@ static void spice_channel_write(SpiceChannel *channel, const void *data, size_t
         spice_channel_flush_wire(channel, data, len);
 }
 
+/* coroutine context */
+static void spice_channel_write_msg(SpiceChannel *channel, SpiceMsgOut *out)
+{
+    uint8_t *data;
+    int free_data;
+    size_t len;
+    uint32_t msg_size;
+
+    g_return_if_fail(channel != NULL);
+    g_return_if_fail(out != NULL);
+    g_return_if_fail(channel == out->channel);
+
+    if (out->ro_check &&
+        spice_channel_get_read_only(channel)) {
+        g_warning("Try to send message while read-only. Please report a bug.");
+        return;
+    }
+
+    msg_size = spice_marshaller_get_total_size(out->marshaller) -
+               spice_header_get_header_size(channel->priv->use_mini_header);
+    spice_header_set_msg_size(out->header, channel->priv->use_mini_header, msg_size);
+    data = spice_marshaller_linearize(out->marshaller, 0, &len, &free_data);
+    /* spice_msg_out_hexdump(out, data, len); */
+    spice_channel_write(channel, data, len);
+
+    if (free_data)
+        free(data);
+
+    spice_msg_out_unref(out);
+}
+
 /*
  * Read at least 1 more byte of data straight off the wire
  * into the requested buffer.
@@ -698,14 +906,7 @@ reread:
 
     if (ret == -1) {
         if (cond != 0) {
-            if (c->wait_interruptible) {
-                if (!g_io_wait_interruptible(&c->wait, c->sock, cond)) {
-                    // SPICE_DEBUG("Read blocking interrupted %d", priv->has_error);
-                    return -EAGAIN;
-                }
-            } else {
-                g_io_wait(c->sock, cond);
-            }
+            g_coroutine_socket_wait(&c->coroutine, c->sock, cond);
             goto reread;
         } else {
             c->has_error = TRUE;
@@ -960,6 +1161,7 @@ static void spice_channel_recv_link_hdr(SpiceChannel *channel)
         goto error;
     }
 
+    SPICE_DEBUG("Peer version: %d:%d", c->peer_hdr.major_version, c->peer_hdr.minor_version);
     if (c->peer_hdr.major_version != c->link_hdr.major_version) {
         if (c->peer_hdr.major_version == 1) {
             /* enter spice 0.4 mode */
@@ -975,6 +1177,11 @@ static void spice_channel_recv_link_hdr(SpiceChannel *channel)
     }
 
     c->peer_msg = spice_malloc(c->peer_hdr.size);
+    if (c->peer_msg == NULL) {
+        g_critical("invalid peer header size: %u", c->peer_hdr.size);
+        goto error;
+    }
+
     c->state = SPICE_CHANNEL_STATE_LINK_MSG;
     return;
 
@@ -1493,7 +1700,9 @@ static void spice_channel_recv_link_msg(SpiceChannel *channel)
             goto error;
         }
     }
-
+    c->use_mini_header = spice_channel_test_common_capability(channel,
+                                                              SPICE_COMMON_CAP_MINI_HEADER);
+    SPICE_DEBUG("use mini header: %d", c->use_mini_header);
     return;
 
 error:
@@ -1502,55 +1711,21 @@ error:
 }
 
 /* system context */
-static void spice_channel_buffered_write(SpiceChannel *channel, const void *data, size_t size)
-{
-    SpiceChannelPrivate *c = channel->priv;
-    size_t left;
-
-    left = c->xmit_buffer_capacity - c->xmit_buffer_size;
-    if (left < size) {
-        c->xmit_buffer_capacity += size + 4095;
-        c->xmit_buffer_capacity &= ~4095;
-
-        c->xmit_buffer = g_realloc(c->xmit_buffer, c->xmit_buffer_capacity);
-    }
-
-    memcpy(&c->xmit_buffer[c->xmit_buffer_size], data, size);
-
-    c->xmit_buffer_size += size;
-}
-
-/* system context */
-/* TODO: we currently flush/wakeup immediately all buffered messages */
 G_GNUC_INTERNAL
-void spice_channel_wakeup(SpiceChannel *channel)
+void spice_channel_wakeup(SpiceChannel *channel, gboolean cancel)
 {
-    SpiceChannelPrivate *c = channel->priv;
+    GCoroutine *c = &channel->priv->coroutine;
 
-    g_io_wakeup(&c->wait);
+    if (cancel)
+        g_coroutine_condition_cancel(c);
+
+    g_coroutine_wakeup(c);
 }
 
-/* coroutine context if @buffered is FALSE,
-   system context if @buffered is TRUE */
-static void spice_channel_send_msg(SpiceChannel *channel, SpiceMsgOut *out, gboolean buffered)
+G_GNUC_INTERNAL
+gboolean spice_channel_get_read_only(SpiceChannel *channel)
 {
-    uint8_t *data;
-    int free_data;
-    size_t len;
-
-    g_return_if_fail(channel != NULL);
-    g_return_if_fail(out != NULL);
-
-    data = spice_marshaller_linearize(out->marshaller, 0,
-                                      &len, &free_data);
-    /* spice_msg_out_hexdump(out, data, len); */
-    if (buffered)
-        spice_channel_buffered_write(channel, data, len);
-    else
-        spice_channel_write(channel, data, len);
-    if (free_data) {
-        free(data);
-    }
+    return spice_session_get_read_only(channel->priv->session);
 }
 
 /* coroutine context */
@@ -1560,54 +1735,65 @@ void spice_channel_recv_msg(SpiceChannel *channel,
 {
     SpiceChannelPrivate *c = channel->priv;
     SpiceMsgIn *in;
+    int header_size;
+    int msg_size;
+    int msg_type;
+    int sub_list_offset = 0;
     int rc;
 
     if (!c->msg_in) {
         c->msg_in = spice_msg_in_new(channel);
     }
     in = c->msg_in;
+    header_size = spice_header_get_header_size(c->use_mini_header);
 
     /* receive message */
-    if (in->hpos < sizeof(in->header)) {
-        rc = spice_channel_read(channel, (uint8_t*)&in->header + in->hpos,
-                                sizeof(in->header) - in->hpos);
+    if (in->hpos < header_size) {
+        rc = spice_channel_read(channel, in->header + in->hpos,
+                                header_size - in->hpos);
         if (rc < 0) {
             g_critical("recv hdr: %s", strerror(errno));
             return;
         }
         in->hpos += rc;
-        if (in->hpos < sizeof(in->header))
+        if (in->hpos < header_size)
             return;
-        in->data = spice_malloc(in->header.size);
+        in->data = spice_malloc(spice_header_get_msg_size(in->header, c->use_mini_header));
     }
-    if (in->dpos < in->header.size) {
+    msg_size = spice_header_get_msg_size(in->header, c->use_mini_header);
+    if (in->dpos < msg_size) {
         rc = spice_channel_read(channel, in->data + in->dpos,
-                                in->header.size - in->dpos);
+                                msg_size - in->dpos);
         if (rc < 0) {
             g_critical("recv msg: %s", strerror(errno));
             return;
         }
         in->dpos += rc;
-        if (in->dpos < in->header.size)
+        if (in->dpos < msg_size)
             return;
     }
 
-    if (in->header.sub_list) {
+    msg_type = spice_header_get_msg_type(in->header, c->use_mini_header);
+    sub_list_offset = spice_header_get_msg_sub_list(in->header, c->use_mini_header);
+
+    if (msg_type == SPICE_MSG_LIST || sub_list_offset) {
         SpiceSubMessageList *sub_list;
         SpiceSubMessage *sub;
         SpiceMsgIn *sub_in;
         int i;
 
-        sub_list = (SpiceSubMessageList *)(in->data + in->header.sub_list);
+        sub_list = (SpiceSubMessageList *)(in->data + sub_list_offset);
         for (i = 0; i < sub_list->size; i++) {
             sub = (SpiceSubMessage *)(in->data + sub_list->sub_messages[i]);
             sub_in = spice_msg_in_sub_new(channel, in, sub);
             sub_in->parsed = c->parser(sub_in->data, sub_in->data + sub_in->dpos,
-                                       sub_in->header.type, c->peer_hdr.minor_version,
+                                       spice_header_get_msg_type(sub_in->header,
+                                                                 c->use_mini_header),
+                                       c->peer_hdr.minor_version,
                                        &sub_in->psize, &sub_in->pfree);
             if (sub_in->parsed == NULL) {
                 g_critical("failed to parse sub-message: %s type %d",
-                           c->name, sub_in->header.type);
+                           c->name, spice_header_get_msg_type(sub_in->header, c->use_mini_header));
                 return;
             }
             msg_handler(channel, sub_in, data);
@@ -1621,17 +1807,20 @@ void spice_channel_recv_msg(SpiceChannel *channel,
         if (!c->message_ack_count) {
             SpiceMsgOut *out = spice_msg_out_new(channel, SPICE_MSGC_ACK);
             spice_msg_out_send_internal(out);
-            spice_msg_out_unref(out);
             c->message_ack_count = c->message_ack_window;
         }
     }
 
+    if (msg_type == SPICE_MSG_LIST) {
+        goto end;
+    }
+
     /* parse message */
-    in->parsed = c->parser(in->data, in->data + in->dpos, in->header.type,
+    in->parsed = c->parser(in->data, in->data + in->dpos, msg_type,
                            c->peer_hdr.minor_version, &in->psize, &in->pfree);
     if (in->parsed == NULL) {
         g_critical("failed to parse message: %s type %d",
-                   c->name, in->header.type);
+                   c->name, msg_type);
         goto end;
     }
 
@@ -1641,6 +1830,10 @@ void spice_channel_recv_msg(SpiceChannel *channel,
     msg_handler(channel, in, data);
 
 end:
+    /* If the server uses full header, the serial is not necessarily equal
+     * to c->in_serial (the server can sometimes skip serials) */
+    c->last_message_serial = spice_header_get_in_msg_serial(in);
+    c->in_serial++;
     /* release message */
     c->msg_in = NULL;
     spice_msg_in_unref(in);
@@ -1657,7 +1850,8 @@ const gchar* spice_channel_type_to_string(gint type)
         [ SPICE_CHANNEL_PLAYBACK ] = "playback",
         [ SPICE_CHANNEL_RECORD ] = "record",
         [ SPICE_CHANNEL_TUNNEL ] = "tunnel",
-        [ SPICE_CHANNEL_SMARTCARD ] = "smartcard"
+        [ SPICE_CHANNEL_SMARTCARD ] = "smartcard",
+        [ SPICE_CHANNEL_USBREDIR ] = "usbredir",
     };
     const char *str = NULL;
 
@@ -1699,11 +1893,17 @@ SpiceChannel *spice_channel_new(SpiceSession *s, int type, int id)
         gtype = SPICE_TYPE_INPUTS_CHANNEL;
         break;
     case SPICE_CHANNEL_PLAYBACK:
-        gtype = SPICE_TYPE_PLAYBACK_CHANNEL;
-        break;
-    case SPICE_CHANNEL_RECORD:
-        gtype = SPICE_TYPE_RECORD_CHANNEL;
+    case SPICE_CHANNEL_RECORD: {
+        gboolean enabled;
+        g_object_get(G_OBJECT(s), "enable-audio", &enabled, NULL);
+        if (!enabled) {
+            g_debug("audio channel is disabled, not creating it");
+            return NULL;
+        }
+        gtype = type == SPICE_CHANNEL_RECORD ?
+            SPICE_TYPE_RECORD_CHANNEL : SPICE_TYPE_PLAYBACK_CHANNEL;
         break;
+    }
 #ifdef USE_SMARTCARD
     case SPICE_CHANNEL_SMARTCARD: {
         gboolean enabled;
@@ -1715,6 +1915,18 @@ SpiceChannel *spice_channel_new(SpiceSession *s, int type, int id)
         gtype = SPICE_TYPE_SMARTCARD_CHANNEL;
         break;
     }
+#endif
+#ifdef USE_USBREDIR
+    case SPICE_CHANNEL_USBREDIR: {
+        gboolean enabled;
+        g_object_get(G_OBJECT(s), "enable-usbredir", &enabled, NULL);
+        if (!enabled) {
+            g_debug("usbredir channel is disabled, not creating it");
+            return NULL;
+        }
+        gtype = SPICE_TYPE_USBREDIR_CHANNEL;
+        break;
+    }
 #endif
     default:
         g_debug("unsupported channel kind: %s: %d",
@@ -1749,11 +1961,15 @@ void spice_channel_destroy(SpiceChannel *channel)
 static void spice_channel_iterate_write(SpiceChannel *channel)
 {
     SpiceChannelPrivate *c = channel->priv;
+    SpiceMsgOut *out;
 
-    if (c->xmit_buffer_size) {
-        spice_channel_write(channel, c->xmit_buffer, c->xmit_buffer_size);
-        c->xmit_buffer_size = 0;
-    }
+    do {
+        g_static_mutex_lock(&c->xmit_queue_lock);
+        out = g_queue_pop_head(&c->xmit_queue);
+        g_static_mutex_unlock(&c->xmit_queue_lock);
+        if (out)
+            spice_channel_write_msg(channel, out);
+    } while (out);
 }
 
 /* coroutine context */
@@ -1776,11 +1992,26 @@ static void spice_channel_iterate_read(SpiceChannel *channel)
         spice_channel_recv_msg(channel,
             (handler_msg_in)SPICE_CHANNEL_GET_CLASS(channel)->handle_msg, NULL);
         break;
+    case SPICE_CHANNEL_STATE_MIGRATING:
+        return;
     default:
         g_critical("unknown state %d", c->state);
     }
 }
 
+static gboolean wait_migration(gpointer data)
+{
+    SpiceChannel *channel = SPICE_CHANNEL(data);
+    SpiceChannelPrivate *c = channel->priv;
+
+    if (c->state != SPICE_CHANNEL_STATE_MIGRATING) {
+        SPICE_DEBUG("unfreeze channel %s", c->name);
+        return TRUE;
+    }
+
+    return FALSE;
+}
+
 /* coroutine context */
 static gboolean spice_channel_iterate(SpiceChannel *channel)
 {
@@ -1788,11 +2019,9 @@ static gboolean spice_channel_iterate(SpiceChannel *channel)
     GIOCondition ret;
 
     do {
-        while (c->state == SPICE_CHANNEL_STATE_MIGRATING) {
-            /* freeze coroutine */
-            coroutine_yield(NULL);
-            g_return_val_if_fail(c->state != SPICE_CHANNEL_STATE_MIGRATING, FALSE);
-        }
+        if (c->state == SPICE_CHANNEL_STATE_MIGRATING &&
+            !g_coroutine_condition_wait(&c->coroutine, wait_migration, channel))
+                SPICE_DEBUG("migration wait cancelled");
 
         if (c->has_error) {
             SPICE_DEBUG("channel has error, breaking loop");
@@ -1800,7 +2029,8 @@ static gboolean spice_channel_iterate(SpiceChannel *channel)
         }
 
         SPICE_CHANNEL_GET_CLASS(channel)->iterate_write(channel);
-        ret = g_io_wait_interruptible(&c->wait, c->sock, G_IO_IN);
+        ret = g_coroutine_socket_wait(&c->coroutine, c->sock, G_IO_IN);
+
 #ifdef WIN32
         /* FIXME: windows gsocket is buggy, it doesn't return correct condition... */
         ret = g_socket_condition_check(c->sock, G_IO_IN);
@@ -1837,7 +2067,7 @@ static gboolean spice_channel_delayed_unref(gpointer data)
     g_return_val_if_fail(channel != NULL, FALSE);
     SPICE_DEBUG("Delayed unref channel %s %p", c->name, channel);
 
-    g_return_val_if_fail(c->coroutine.exited == TRUE, FALSE);
+    g_return_val_if_fail(c->coroutine.coroutine.exited == TRUE, FALSE);
 
     g_object_unref(G_OBJECT(data));
 
@@ -1850,6 +2080,7 @@ static void *spice_channel_coroutine(void *data)
     SpiceChannel *channel = SPICE_CHANNEL(data);
     SpiceChannelPrivate *c = channel->priv;
     guint verify;
+    int rc, delay_val = 1;
 
     SPICE_DEBUG("Started background coroutine %p for %s", &c->coroutine, c->name);
 
@@ -1869,7 +2100,7 @@ static void *spice_channel_coroutine(void *data)
     }
 
 reconnect:
-    c->sock = spice_session_channel_open_host(c->session, c->tls);
+    c->sock = spice_session_channel_open_host(c->session, channel, c->tls);
     if (c->sock == NULL) {
         if (!c->tls) {
             SPICE_DEBUG("connection failed, trying with TLS port");
@@ -1885,8 +2116,6 @@ reconnect:
     c->has_error = FALSE;
 
     if (c->tls) {
-        int rc;
-
         c->ctx = SSL_CTX_new(TLSv1_method());
         if (c->ctx == NULL) {
             g_critical("SSL_CTX_new failed");
@@ -1950,7 +2179,7 @@ ssl_reconnect:
         if (rc <= 0) {
             rc = SSL_get_error(c->ssl, rc);
             if (rc == SSL_ERROR_WANT_READ || rc == SSL_ERROR_WANT_WRITE) {
-                g_io_wait(c->sock, G_IO_OUT|G_IO_ERR|G_IO_HUP);
+                g_coroutine_socket_wait(&c->coroutine, c->sock, G_IO_OUT|G_IO_ERR|G_IO_HUP);
                 goto ssl_reconnect;
             } else {
                 g_warning("%s: SSL_connect: %s",
@@ -1962,6 +2191,13 @@ ssl_reconnect:
     }
 
 connected:
+    rc = setsockopt(g_socket_get_fd(c->sock), IPPROTO_TCP, TCP_NODELAY,
+                    (const char*)&delay_val, sizeof(delay_val));
+    if (rc != 0) {
+        g_warning("%s: could not set sockopt TCP_NODELAY: %s", c->name,
+                  strerror(errno));
+    }
+
     c->state = SPICE_CHANNEL_STATE_LINK_HDR;
     spice_channel_send_link(channel);
 
@@ -1998,7 +2234,7 @@ static gboolean connect_delayed(gpointer data)
     SPICE_DEBUG("Open coroutine starting %p", channel);
     c->connect_delayed_id = 0;
 
-    co = &c->coroutine;
+    co = &c->coroutine.coroutine;
 
     co->stack_size = 16 << 20; /* 16Mb */
     co->entry = spice_channel_coroutine;
@@ -2033,6 +2269,7 @@ static gboolean channel_connect(SpiceChannel *channel)
         }
     }
     c->state = SPICE_CHANNEL_STATE_CONNECTING;
+    c->xmit_queue_blocked = FALSE;
 
     g_return_val_if_fail(c->sock == NULL, FALSE);
     g_object_ref(G_OBJECT(channel)); /* Unref'd when co-routine exits */
@@ -2084,18 +2321,10 @@ gboolean spice_channel_open_fd(SpiceChannel *channel, int fd)
 }
 
 /* system or coroutine context */
-static void channel_disconnect(SpiceChannel *channel)
+static void channel_reset(SpiceChannel *channel, gboolean migrating)
 {
     SpiceChannelPrivate *c = SPICE_CHANNEL_GET_PRIVATE(channel);
 
-    g_return_if_fail(c != NULL);
-
-    if (c->state == SPICE_CHANNEL_STATE_UNCONNECTED) {
-        return;
-    }
-
-    c->has_error = TRUE; /* break the loop */
-
     if (c->connect_delayed_id) {
         g_source_remove(c->connect_delayed_id);
         c->connect_delayed_id = 0;
@@ -2133,20 +2362,49 @@ static void channel_disconnect(SpiceChannel *channel)
     c->peer_msg = NULL;
     c->peer_pos = 0;
 
-    if (c->xmit_buffer) {
-        g_free(c->xmit_buffer);
-        c->xmit_buffer = NULL;
-        c->xmit_buffer_size = 0;
-        c->xmit_buffer_capacity = 0;
+    g_static_mutex_lock(&c->xmit_queue_lock);
+    c->xmit_queue_blocked = TRUE; /* Disallow queuing new messages */
+    g_queue_foreach(&c->xmit_queue, (GFunc)spice_msg_out_unref, NULL);
+    g_queue_clear(&c->xmit_queue);
+    if (c->xmit_queue_wakeup_id) {
+        g_source_remove(c->xmit_queue_wakeup_id);
+        c->xmit_queue_wakeup_id = 0;
     }
+    g_static_mutex_unlock(&c->xmit_queue_lock);
 
     g_array_set_size(c->remote_common_caps, 0);
     g_array_set_size(c->remote_caps, 0);
     g_array_set_size(c->common_caps, 0);
     g_array_set_size(c->caps, 0);
+    /* Restore our default capabilities in case the channel gets re-used */
+    spice_channel_set_common_capability(channel, SPICE_COMMON_CAP_PROTOCOL_AUTH_SELECTION);
+    spice_channel_set_common_capability(channel, SPICE_COMMON_CAP_MINI_HEADER);
+}
+
+/* system or coroutine context */
+G_GNUC_INTERNAL
+void spice_channel_reset(SpiceChannel *channel, gboolean migrating)
+{
+    SPICE_CHANNEL_GET_CLASS(channel)->channel_reset(channel, migrating);
+}
+
+/* system or coroutine context */
+static void channel_disconnect(SpiceChannel *channel)
+{
+    SpiceChannelPrivate *c = SPICE_CHANNEL_GET_PRIVATE(channel);
+
+    g_return_if_fail(c != NULL);
+
+    if (c->state == SPICE_CHANNEL_STATE_UNCONNECTED)
+        return;
+
+    c->has_error = TRUE; /* break the loop */
+
+    spice_channel_reset(channel, FALSE);
 
     if (c->state == SPICE_CHANNEL_STATE_READY)
         emit_main_context(channel, SPICE_CHANNEL_EVENT, SPICE_CHANNEL_CLOSED);
+
     g_return_if_fail(SPICE_IS_CHANNEL(channel));
     c->state = SPICE_CHANNEL_STATE_UNCONNECTED;
 }
@@ -2164,6 +2422,7 @@ void spice_channel_disconnect(SpiceChannel *channel, SpiceChannelEvent reason)
 {
     SpiceChannelPrivate *c = SPICE_CHANNEL_GET_PRIVATE(channel);
 
+    SPICE_DEBUG("channel disconnect %d", reason);
     g_return_if_fail(c != NULL);
 
     if (c->state == SPICE_CHANNEL_STATE_UNCONNECTED)
@@ -2176,13 +2435,11 @@ void spice_channel_disconnect(SpiceChannel *channel, SpiceChannelEvent reason)
 
     if (c->state == SPICE_CHANNEL_STATE_MIGRATING) {
         c->state = SPICE_CHANNEL_STATE_READY;
-        coroutine_yieldto(&c->coroutine, NULL);
     } else
-        spice_channel_wakeup(channel);
+        spice_channel_wakeup(channel, TRUE);
 
-    if (reason != SPICE_CHANNEL_NONE) {
+    if (reason != SPICE_CHANNEL_NONE)
         g_signal_emit(G_OBJECT(channel), signals[SPICE_CHANNEL_EVENT], 0, reason);
-    }
 }
 
 static gboolean test_capability(GArray *caps, guint32 cap)
@@ -2308,16 +2565,25 @@ void spice_channel_swap(SpiceChannel *channel, SpiceChannel *swap)
         SSL_CTX *ctx = c->ctx;
         SSL *ssl = c->ssl;
         SpiceOpenSSLVerify *sslverify = c->sslverify;
+        uint64_t in_serial = c->in_serial;
+        uint64_t out_serial = c->out_serial;
+        gboolean use_mini_header = c->use_mini_header;
 
         c->sock = s->sock;
         c->ctx = s->ctx;
         c->ssl = s->ssl;
         c->sslverify = s->sslverify;
+        c->in_serial = s->in_serial;
+        c->out_serial = s->out_serial;
+        c->use_mini_header = s->use_mini_header;
 
         s->sock = sock;
         s->ctx = ctx;
         s->ssl = ssl;
         s->sslverify = sslverify;
+        s->in_serial = in_serial;
+        s->out_serial = out_serial;
+        s->use_mini_header = use_mini_header;
     }
 
 #if HAVE_SASL
