@@ -28,57 +28,13 @@
 #include "ring.h"
 
 #include "gio-coroutine.h"
+#include "glib-compat.h"
 
 struct channel {
     SpiceChannel      *channel;
     RingItem          link;
 };
 
-struct _SpiceSessionPrivate {
-    char              *host;
-    char              *port;
-    char              *tls_port;
-    char              *password;
-    char              *ca_file;
-    char              *ciphers;
-    GByteArray        *pubkey;
-    char              *cert_subject;
-    guint             verify;
-
-    /* whether to enable smartcard event forwarding to the server */
-    gboolean          smartcard;
-
-    /* list of certificates to use for the software smartcard reader if
-     * enabled. For now, it has to contain exactly 3 certificates for
-     * the software reader to be functional
-     */
-    GStrv             smartcard_certificates;
-
-    /* path to the local certificate database to use to lookup the
-     * certificates stored in 'certificates'. If NULL, libcacard will
-     * fallback to using a default database.
-     */
-    char *            smartcard_db;
-    GStrv             disable_effects;
-    gint              color_depth;
-
-    int               connection_id;
-    int               protocol;
-    SpiceChannel      *cmain; /* weak reference */
-    Ring              channels;
-    guint32           mm_time;
-    gboolean          client_provided_sockets;
-    guint64           mm_time_at_clock;
-    SpiceSession      *migration;
-    GList             *migration_left;
-    SpiceSessionMigration migration_state;
-    gboolean          disconnecting;
-
-    display_cache     images;
-    display_cache     palettes;
-    SpiceGlzDecoderWindow *glz_window;
-};
-
 /**
  * SECTION:spice-session
  * @short_description: handles connection details, and active channels
@@ -136,11 +92,17 @@ enum {
     PROP_CERT_SUBJECT,
     PROP_VERIFY,
     PROP_MIGRATION_STATE,
+    PROP_AUDIO,
     PROP_SMARTCARD,
     PROP_SMARTCARD_CERTIFICATES,
     PROP_SMARTCARD_DB,
+    PROP_USBREDIR,
+    PROP_INHIBIT_KEYBOARD_GRAB,
     PROP_DISABLE_EFFECTS,
     PROP_COLOR_DEPTH,
+    PROP_READ_ONLY,
+    PROP_CACHE_SIZE,
+    PROP_GLZ_WINDOW_SIZE,
 };
 
 /* signals */
@@ -159,7 +121,6 @@ static void spice_session_init(SpiceSession *session)
 
     SPICE_DEBUG("New session (compiled from package " PACKAGE_STRING ")");
     s = session->priv = SPICE_SESSION_GET_PRIVATE(session);
-    memset(s, 0, sizeof(*s));
 
     ring_init(&s->channels);
     cache_init(&s->images, "image");
@@ -188,6 +149,15 @@ spice_session_dispose(GObject *gobject)
         s->migration_left = NULL;
     }
 
+    if (s->after_main_init) {
+        g_source_remove(s->after_main_init);
+        s->after_main_init = 0;
+    }
+
+    g_clear_object(&s->audio_manager);
+    g_clear_object(&s->gtk_session);
+    g_clear_object(&s->usb_manager);
+
     /* Chain up to the parent class */
     if (G_OBJECT_CLASS(spice_session_parent_class)->dispose)
         G_OBJECT_CLASS(spice_session_parent_class)->dispose(gobject);
@@ -274,6 +244,8 @@ static int spice_uri_parse(SpiceSession *session, const char *original_uri)
     SpiceSessionPrivate *s = SPICE_SESSION_GET_PRIVATE(session);
     char host[128], key[32], value[128];
     char *port = NULL, *tls_port = NULL, *uri = NULL, *password = NULL;
+    char **target_key;
+    int punctuation = 0;
     int len, pos = 0;
 
     g_return_val_if_fail(original_uri != NULL, -1);
@@ -284,28 +256,55 @@ static int spice_uri_parse(SpiceSession *session, const char *original_uri)
     if (sscanf(uri, "spice://%127[-.0-9a-zA-Z]%n", host, &len) != 1)
         goto fail;
     pos += len;
+
+    if (uri[pos] == '/')
+        pos++;
+
     for (;;) {
         if (uri[pos] == '?' || uri[pos] == ';' || uri[pos] == '&') {
             pos++;
+            punctuation++;
             continue;
         }
         if (uri[pos] == 0) {
             break;
         }
-        if (sscanf(uri + pos, "%31[-a-zA-Z0-9]=%127[^;&]%n", key, value, &len) != 2)
-            goto fail;
-        pos += len;
-        if (g_str_equal(key, "port")) {
+        if (uri[pos] == ':') {
+            if (punctuation++) {
+                g_warning("colon seen after a previous punctuation (?;&:)");
+                goto fail;
+            }
+            pos++;
+            /* port numbers are 16 bit, fits in five decimal figures. */
+            if (sscanf(uri + pos, "%5[0-9]%n", value, &len) != 1)
+                goto fail;
             port = g_strdup(value);
+            pos += len;
+            continue;
+        } else {
+            if (sscanf(uri + pos, "%31[-a-zA-Z0-9]=%127[^;&]%n", key, value, &len) != 2)
+                goto fail;
+        }
+        pos += len;
+        target_key = NULL;
+        if (g_str_equal(key, "port")) {
+            target_key = &port;
         } else if (g_str_equal(key, "tls-port")) {
-            tls_port = g_strdup(value);
+            target_key = &tls_port;
         } else if (g_str_equal(key, "password")) {
-            password = g_strdup(value);
+            target_key = &password;
             g_warning("password may be visible in process listings");
         } else {
             g_warning("unknown key in spice URI parsing: %s", key);
             goto fail;
         }
+        if (target_key) {
+            if (*target_key) {
+                g_warning("double set of %s", key);
+                goto fail;
+            }
+            *target_key = g_strdup(value);
+        }
     }
 
     if (port == NULL && tls_port == NULL) {
@@ -393,12 +392,30 @@ static void spice_session_get_property(GObject    *gobject,
     case PROP_SMARTCARD_DB:
         g_value_set_string(value, s->smartcard_db);
         break;
+    case PROP_USBREDIR:
+        g_value_set_boolean(value, s->usbredir);
+        break;
+    case PROP_INHIBIT_KEYBOARD_GRAB:
+        g_value_set_boolean(value, s->inhibit_keyboard_grab);
+        break;
     case PROP_DISABLE_EFFECTS:
         g_value_set_boxed(value, s->disable_effects);
         break;
     case PROP_COLOR_DEPTH:
         g_value_set_int(value, s->color_depth);
         break;
+    case PROP_AUDIO:
+        g_value_set_boolean(value, s->audio);
+        break;
+    case PROP_READ_ONLY:
+        g_value_set_boolean(value, s->read_only);
+        break;
+    case PROP_CACHE_SIZE:
+        g_value_set_int(value, s->images_cache_size);
+        break;
+    case PROP_GLZ_WINDOW_SIZE:
+        g_value_set_int(value, s->glz_window_size);
+        break;
     default:
 	G_OBJECT_WARN_INVALID_PROPERTY_ID(gobject, prop_id, pspec);
 	break;
@@ -479,6 +496,12 @@ static void spice_session_set_property(GObject      *gobject,
         g_free(s->smartcard_db);
         s->smartcard_db = g_value_dup_string(value);
         break;
+    case PROP_USBREDIR:
+        s->usbredir = g_value_get_boolean(value);
+        break;
+    case PROP_INHIBIT_KEYBOARD_GRAB:
+        s->inhibit_keyboard_grab = g_value_get_boolean(value);
+        break;
     case PROP_DISABLE_EFFECTS:
         g_strfreev(s->disable_effects);
         s->disable_effects = g_value_dup_boxed(value);
@@ -486,6 +509,19 @@ static void spice_session_set_property(GObject      *gobject,
     case PROP_COLOR_DEPTH:
         s->color_depth = g_value_get_int(value);
         break;
+    case PROP_AUDIO:
+        s->audio = g_value_get_boolean(value);
+        break;
+    case PROP_READ_ONLY:
+        s->read_only = g_value_get_boolean(value);
+        g_object_notify(gobject, "read-only");
+        break;
+    case PROP_CACHE_SIZE:
+        s->images_cache_size = g_value_get_int(value);
+        break;
+    case PROP_GLZ_WINDOW_SIZE:
+        s->glz_window_size = g_value_get_int(value);
+        break;
     default:
         G_OBJECT_WARN_INVALID_PROPERTY_ID(gobject, prop_id, pspec);
         break;
@@ -748,6 +784,23 @@ static void spice_session_class_init(SpiceSessionClass *klass)
                           G_PARAM_READWRITE |
                           G_PARAM_STATIC_STRINGS));
 
+    /**
+     * SpiceSession:enable-audio:
+     *
+     * If set to TRUE, the audio channels will be enabled for
+     * playback and recording.
+     *
+     * Since: 0.8
+     **/
+    g_object_class_install_property
+        (gobject_class, PROP_AUDIO,
+         g_param_spec_boolean("enable-audio",
+                          "Enable audio channels",
+                          "Enable audio channels",
+                          TRUE,
+                          G_PARAM_READWRITE | G_PARAM_CONSTRUCT |
+                          G_PARAM_STATIC_STRINGS));
+
     /**
      * SpiceSession:smartcard-certificates:
      *
@@ -787,6 +840,41 @@ static void spice_session_class_init(SpiceSessionClass *klass)
                               G_PARAM_WRITABLE |
                               G_PARAM_STATIC_STRINGS));
 
+    /**
+     * SpiceSession:enable-usbredir:
+     *
+     * If set to TRUE, the usbredir channel will be enabled and USB devices
+     * can be redirected to the guest
+     *
+     * Since: 0.8
+     **/
+    g_object_class_install_property
+        (gobject_class, PROP_USBREDIR,
+         g_param_spec_boolean("enable-usbredir",
+                          "Enable USB device redirection",
+                          "Forward USB devices to the SPICE server",
+                          TRUE,
+                          G_PARAM_READWRITE | G_PARAM_CONSTRUCT |
+                          G_PARAM_STATIC_STRINGS));
+
+    /**
+     * SpiceSession::inhibit-keyboard-grab
+     *
+     * This boolean is set by the usbredir channel to indicate to #SpiceDisplay
+     * that the keyboard grab should be temporarily released, because it is
+     * going to invoke policykit. It will get reset when the usbredir channel
+     * is done with polickit.
+     *
+     * Since: 0.8
+     **/
+    g_object_class_install_property
+        (gobject_class, PROP_INHIBIT_KEYBOARD_GRAB,
+         g_param_spec_boolean("inhibit-keyboard-grab",
+                        "Inhibit Keyboard Grab",
+                        "Request that SpiceDisplays don't grab the keyboard",
+                        FALSE,
+                        G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+
     /**
      * SpiceSession::channel-new:
      * @session: the session that emitted the signal
@@ -823,6 +911,52 @@ static void spice_session_class_init(SpiceSessionClass *klass)
                      1,
                      SPICE_TYPE_CHANNEL);
 
+    /**
+     * SpiceSession:read-only:
+     *
+     * Whether this connection is read-only mode.
+     *
+     * Since: 0.8
+     **/
+    g_object_class_install_property
+        (gobject_class, PROP_READ_ONLY,
+         g_param_spec_boolean("read-only", "Read-only",
+                              "Whether this connection is read-only mode",
+                              FALSE,
+                              G_PARAM_READWRITE |
+                              G_PARAM_CONSTRUCT |
+                              G_PARAM_STATIC_STRINGS));
+
+    /**
+     * SpiceSession:cache-size:
+     *
+     * Images cache size. If 0, don't set.
+     *
+     **/
+    g_object_class_install_property
+        (gobject_class, PROP_CACHE_SIZE,
+         g_param_spec_int("cache-size",
+                          "Cache size",
+                          "Images cache size (bytes)",
+                          0, G_MAXINT, 0,
+                          G_PARAM_READWRITE |
+                          G_PARAM_STATIC_STRINGS));
+
+    /**
+     * SpiceSession:glz-window-size:
+     *
+     * Glz window size. If 0, don't set.
+     *
+     **/
+    g_object_class_install_property
+        (gobject_class, PROP_GLZ_WINDOW_SIZE,
+         g_param_spec_int("glz-window-size",
+                          "Glz window size",
+                          "Glz window size (bytes)",
+                          0, LZ_MAX_WINDOW_SIZE * 4, 0,
+                          G_PARAM_READWRITE |
+                          G_PARAM_STATIC_STRINGS));
+
     g_type_class_add_private(klass, sizeof(SpiceSessionPrivate));
 }
 
@@ -997,8 +1131,7 @@ void spice_session_set_migration(SpiceSession *session, SpiceSession *migration)
 }
 
 G_GNUC_INTERNAL
-SpiceChannel* get_channel_by_id_and_type(SpiceSession *session,
-                                         gint id, gint type)
+SpiceChannel* spice_session_lookup_channel(SpiceSession *session, gint id, gint type)
 {
     RingItem *ring, *next;
     SpiceSessionPrivate *s = session->priv;
@@ -1043,9 +1176,9 @@ void spice_session_abort_migration(SpiceSession *session)
             continue;
 
         spice_channel_swap(c->channel,
-            get_channel_by_id_and_type(s->migration,
-                                       spice_channel_get_channel_id(c->channel),
-                                       spice_channel_get_channel_type(c->channel)));
+            spice_session_lookup_channel(s->migration,
+                                         spice_channel_get_channel_id(c->channel),
+                                         spice_channel_get_channel_type(c->channel)));
     }
 
     g_list_free(s->migration_left);
@@ -1072,7 +1205,7 @@ void spice_session_channel_migrate(SpiceSession *session, SpiceChannel *channel)
     type = spice_channel_get_channel_type(channel);
     SPICE_DEBUG("migrating channel id:%d type:%d", id, type);
 
-    c = get_channel_by_id_and_type(s->migration, id, type);
+    c = spice_session_lookup_channel(s->migration, id, type);
     g_return_if_fail(c != NULL);
 
     spice_channel_swap(channel, c);
@@ -1087,6 +1220,98 @@ void spice_session_channel_migrate(SpiceSession *session, SpiceChannel *channel)
     }
 }
 
+/* main context */
+static gboolean after_main_init(gpointer data)
+{
+    SpiceSession *self = data;
+    SpiceSessionPrivate *s = self->priv;
+    GList *l;
+
+    for (l = s->migration_left; l != NULL; ) {
+        SpiceChannel *channel = l->data;
+        l = l->next;
+
+        spice_session_channel_migrate(self, channel);
+        channel->priv->state = SPICE_CHANNEL_STATE_READY;
+        spice_channel_up(channel);
+    }
+
+    s->after_main_init = 0;
+    return FALSE;
+}
+
+/* coroutine context */
+G_GNUC_INTERNAL
+gboolean spice_session_migrate_after_main_init(SpiceSession *self)
+{
+    SpiceSessionPrivate *s = self->priv;
+
+    if (!s->migrate_wait_init)
+        return FALSE;
+
+    g_return_val_if_fail(g_list_length(s->migration_left) != 0, FALSE);
+    g_return_val_if_fail(s->after_main_init == 0, FALSE);
+
+    s->migrate_wait_init = FALSE;
+    s->after_main_init = g_idle_add(after_main_init, self);
+
+    return TRUE;
+}
+
+/* main context */
+G_GNUC_INTERNAL
+void spice_session_migrate_end(SpiceSession *self)
+{
+    SpiceSessionPrivate *s = self->priv;
+    SpiceMsgOut *out;
+    GList *l;
+
+    g_return_if_fail(s->migration);
+    g_return_if_fail(s->migration->priv->cmain);
+    g_return_if_fail(g_list_length(s->migration_left) != 0);
+
+    /* disconnect and reset all channels */
+    for (l = s->migration_left; l != NULL; ) {
+        SpiceChannel *channel = l->data;
+        l = l->next;
+
+        /* reset for migration, disconnect */
+        spice_channel_reset(channel, TRUE);
+
+        if (SPICE_IS_MAIN_CHANNEL(channel)) {
+            /* migrate main to target, so we can start talking */
+            spice_session_channel_migrate(self, channel);
+        } else {
+            /* freeze other channels */
+            channel->priv->state = SPICE_CHANNEL_STATE_MIGRATING;
+        }
+    }
+
+    spice_session_palettes_clear(self);
+    spice_session_images_clear(self);
+    glz_decoder_window_clear(s->glz_window);
+
+    /* send MIGRATE_END to target */
+    out = spice_msg_out_new(s->cmain, SPICE_MSGC_MAIN_MIGRATE_END);
+    spice_msg_out_send(out);
+
+    /* now wait after main init for the rest of channels migration */
+    s->migrate_wait_init = TRUE;
+}
+
+/**
+ * spice_session_get_read_only:
+ * @session: a #SpiceSession
+ *
+ * Returns: wether the @session is in read-only mode.
+ **/
+gboolean spice_session_get_read_only(SpiceSession *self)
+{
+    g_return_val_if_fail(SPICE_IS_SESSION(self), FALSE);
+
+    return self->priv->read_only;
+}
+
 /**
  * spice_session_disconnect:
  * @session:
@@ -1147,12 +1372,42 @@ GList *spice_session_get_channels(SpiceSession *session)
     return list;
 }
 
+/**
+ * spice_session_has_channel_type:
+ * @session: a #SpiceSession
+ *
+ * See if there is a @type channel in the channels associated with this
+ * @session.
+ *
+ * Returns: TRUE if a @type channel is available otherwise FALSE.
+ **/
+gboolean spice_session_has_channel_type(SpiceSession *session, gint type)
+{
+    SpiceSessionPrivate *s = SPICE_SESSION_GET_PRIVATE(session);
+    struct channel *item;
+    RingItem *ring;
+
+    g_return_val_if_fail(s != NULL, FALSE);
+
+    for (ring = ring_get_head(&s->channels);
+         ring != NULL;
+         ring = ring_next(&s->channels, ring)) {
+        item = SPICE_CONTAINEROF(ring, struct channel, link);
+        if (spice_channel_get_channel_type(item->channel) == type) {
+            return TRUE;
+        }
+    }
+    return FALSE;
+}
+
 /* ------------------------------------------------------------------ */
 /* private functions                                                  */
 
-static GSocket *channel_connect_socket(GSocketAddress *sockaddr,
+static GSocket *channel_connect_socket(SpiceChannel *channel,
+                                       GSocketAddress *sockaddr,
                                        GError **error)
 {
+    SpiceChannelPrivate *c = channel->priv;
     GSocket *sock = g_socket_new(g_socket_address_get_family(sockaddr),
                                  G_SOCKET_TYPE_STREAM,
                                  G_SOCKET_PROTOCOL_DEFAULT,
@@ -1166,7 +1421,7 @@ static GSocket *channel_connect_socket(GSocketAddress *sockaddr,
         if (*error && (*error)->code == G_IO_ERROR_PENDING) {
             g_clear_error(error);
             SPICE_DEBUG("Socket pending");
-            g_io_wait(sock, G_IO_OUT | G_IO_ERR | G_IO_HUP);
+            g_coroutine_socket_wait(&c->coroutine, sock, G_IO_OUT | G_IO_ERR | G_IO_HUP);
 
             if (!g_socket_check_connect_result(sock, error)) {
                 SPICE_DEBUG("Failed to connect %s", (*error)->message);
@@ -1185,8 +1440,10 @@ static GSocket *channel_connect_socket(GSocketAddress *sockaddr,
     return sock;
 }
 
+/* coroutine context */
 G_GNUC_INTERNAL
-GSocket* spice_session_channel_open_host(SpiceSession *session, gboolean use_tls)
+GSocket* spice_session_channel_open_host(SpiceSession *session, SpiceChannel *channel,
+                                         gboolean use_tls)
 {
     SpiceSessionPrivate *s = SPICE_SESSION_GET_PRIVATE(session);
     GSocketConnectable *addr;
@@ -1216,7 +1473,9 @@ GSocket* spice_session_channel_open_host(SpiceSession *session, gboolean use_tls
            (sockaddr = g_socket_address_enumerator_next(enumerator, NULL, &conn_error))) {
         SPICE_DEBUG("Trying one socket");
         g_clear_error(&conn_error);
-        sock = channel_connect_socket(sockaddr, &conn_error);
+        sock = channel_connect_socket(channel, sockaddr, &conn_error);
+        if (conn_error != NULL)
+            SPICE_DEBUG("%s", conn_error->message);
         g_object_unref(sockaddr);
     }
     g_object_unref(enumerator);
@@ -1449,3 +1708,27 @@ void spice_session_get_caches(SpiceSession *session,
     if (glz_window)
         *glz_window = s->glz_window;
 }
+
+G_GNUC_INTERNAL
+void spice_session_set_caches_hints(SpiceSession *session,
+                                    uint32_t pci_ram_size,
+                                    uint32_t display_channels_count)
+{
+    SpiceSessionPrivate *s = SPICE_SESSION_GET_PRIVATE(session);
+
+    g_return_if_fail(s != NULL);
+
+    s->pci_ram_size = pci_ram_size;
+    s->display_channels_count = display_channels_count;
+
+    /* TODO: when setting cache and window size, we should consider the client's
+     *       available memory and the number of display channels */
+    if (s->images_cache_size == 0) {
+        s->images_cache_size = IMAGES_CACHE_SIZE_DEFAULT;
+    }
+
+    if (s->glz_window_size == 0) {
+        s->glz_window_size = MIN(MAX_GLZ_WINDOW_SIZE_DEFAULT, pci_ram_size / 2);
+        s->glz_window_size = MAX(MIN_GLZ_WINDOW_SIZE_DEFAULT, s->glz_window_size);
+    }
+}
