@@ -71,16 +71,10 @@ void spicex_draw_event(SpiceDisplay *display, cairo_t *cr)
 {
     SpiceDisplayPrivate *d = SPICE_DISPLAY_GET_PRIVATE(display);
     int fbw = d->width, fbh = d->height;
-    int mx = 0, my = 0;
     int ww, wh;
 
     gdk_drawable_get_size(gtk_widget_get_window(GTK_WIDGET(display)), &ww, &wh);
 
-    if (ww > fbw)
-        mx = (ww - fbw) / 2;
-    if (wh > fbh)
-        my = (wh - fbh) / 2;
-
     /* If we don't have a pixmap, or we're not scaling, then
        we need to fill with background color */
     if (!d->ximage ||
@@ -92,7 +86,7 @@ void spicex_draw_event(SpiceDisplay *display, cairo_t *cr)
            behaviour of drawing the rectangle from right to left
            to cut out the whole */
         if (d->ximage)
-            cairo_rectangle(cr, mx + fbw, my,
+            cairo_rectangle(cr, d->mx + fbw, d->my,
                             -1 * fbw, fbh);
         cairo_fill(cr);
     }
@@ -101,15 +95,23 @@ void spicex_draw_event(SpiceDisplay *display, cairo_t *cr)
     if (d->ximage) {
         if (d->allow_scaling) {
             double sx, sy;
-            /* Scale to fill window */
-            sx = (double)ww / (double)fbw;
-            sy = (double)wh / (double)fbh;
+            spice_display_get_scaling(display, &sx, &sy);
             cairo_scale(cr, sx, sy);
             cairo_set_source_surface(cr, d->ximage, 0, 0);
         } else {
-            cairo_set_source_surface(cr, d->ximage, mx, my);
+            cairo_set_source_surface(cr, d->ximage, d->mx, d->my);
         }
         cairo_paint(cr);
+
+        if (d->mouse_mode == SPICE_MOUSE_MODE_SERVER) {
+            GdkPixbuf *image = d->mouse_pixbuf;
+            if (image != NULL) {
+                gdk_cairo_set_source_pixbuf(cr, image,
+                                            d->mx + d->mouse_guest_x - d->mouse_hotspot.x,
+                                            d->my + d->mouse_guest_y - d->mouse_hotspot.y);
+                cairo_paint(cr);
+            }
+        }
     }
 }
 
