@@ -1,6 +1,6 @@
 /* -*- Mode: C; c-basic-offset: 4; indent-tabs-mode: nil -*- */
 /*
-   Copyright (C) 2010 Red Hat, Inc.
+   Copyright (C) 2010-2011 Red Hat, Inc.
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
@@ -33,10 +33,12 @@
 #endif
 
 #include "spice-widget.h"
+#include "spice-gtk-session.h"
 #include "spice-audio.h"
 #include "spice-common.h"
 #include "spice-cmdline.h"
 #include "spice-option.h"
+#include "usb-device-widget.h"
 
 /* config */
 static gboolean fullscreen = false;
@@ -74,6 +76,7 @@ struct spice_window {
 
 struct spice_connection {
     SpiceSession     *session;
+    SpiceGtkSession  *gtk_session;
     spice_window     *wins[4];
     SpiceAudio       *audio;
     char             *mouse_state;
@@ -96,6 +99,11 @@ static void connection_disconnect(spice_connection *conn);
 static void connection_destroy(spice_connection *conn);
 static void resolution_fullscreen(struct spice_window *win);
 static void resolution_restore(struct spice_window *win);
+static void usb_connect_failed(GObject               *object,
+                               SpiceUsbDevice        *device,
+                               GError                *error,
+                               gpointer               data);
+static gboolean is_gtk_session_property(const gchar *property);
 
 /* ------------------------------------------------------------------ */
 
@@ -231,7 +239,7 @@ static int connect_dialog(SpiceSession *session)
         }
     }
 
-    label = gtk_label_new("Recent connexions:");
+    label = gtk_label_new("Recent connections:");
     gtk_box_pack_start(GTK_BOX(area), label, TRUE, TRUE, 0);
     gtk_misc_set_alignment(GTK_MISC(label), 0, 0.5);
 #ifndef WIN32
@@ -313,14 +321,14 @@ static void menu_cb_copy(GtkAction *action, void *data)
 {
     struct spice_window *win = data;
 
-    spice_display_copy_to_guest(SPICE_DISPLAY(win->spice));
+    spice_gtk_session_copy_to_guest(win->conn->gtk_session);
 }
 
 static void menu_cb_paste(GtkAction *action, void *data)
 {
     struct spice_window *win = data;
 
-    spice_display_paste_from_guest(SPICE_DISPLAY(win->spice));
+    spice_gtk_session_paste_from_guest(win->conn->gtk_session);
 }
 
 static void window_set_fullscreen(struct spice_window *win, gboolean fs)
@@ -410,17 +418,67 @@ static void menu_cb_remove_smartcard(GtkAction *action, void *data)
 }
 #endif
 
+#ifdef USE_USBREDIR
+static void menu_cb_select_usb_devices(GtkAction *action, void *data)
+{
+    GtkWidget *dialog, *area, *usb_device_widget;
+    struct spice_window *win = data;
+
+    /* Create the widgets */
+    dialog = gtk_dialog_new_with_buttons(
+                    _("Select USB devices for redirection"),
+                    GTK_WINDOW(win->toplevel),
+                    GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
+                    GTK_STOCK_OK, GTK_RESPONSE_ACCEPT,
+                    NULL);
+    gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_ACCEPT);
+    area = gtk_dialog_get_content_area(GTK_DIALOG(dialog));
+
+    usb_device_widget = spice_usb_device_widget_new(win->conn->session,
+                                                    "%s %s");
+    g_signal_connect(usb_device_widget, "connect-failed",
+                     G_CALLBACK(usb_connect_failed), NULL);
+    gtk_box_pack_start(GTK_BOX(area), usb_device_widget, TRUE, TRUE, 5);
+
+    /* show and run */
+    gtk_widget_show_all(dialog);
+    gtk_dialog_run(GTK_DIALOG(dialog));
+    gtk_widget_destroy(dialog);
+}
+#endif
+
 static void menu_cb_bool_prop(GtkToggleAction *action, gpointer data)
 {
     struct spice_window *win = data;
     gboolean state = gtk_toggle_action_get_active(action);
     const char *name;
+    gpointer object;
 
     name = gtk_action_get_name(GTK_ACTION(action));
     SPICE_DEBUG("%s: %s = %s", __FUNCTION__, name, state ? _("yes") : _("no"));
 
     g_key_file_set_boolean(keyfile, "general", name, state);
-    g_object_set(G_OBJECT(win->spice), name, state, NULL);
+
+    if (is_gtk_session_property(name)) {
+        object = win->conn->gtk_session;
+    } else {
+        object = win->spice;
+    }
+    g_object_set(object, name, state, NULL);
+}
+
+static void menu_cb_conn_bool_prop_changed(GObject    *gobject,
+                                           GParamSpec *pspec,
+                                           gpointer    user_data)
+{
+    struct spice_window *win = user_data;
+    const gchar *property = g_param_spec_get_name(pspec);
+    GtkAction *toggle;
+    gboolean state;
+
+    toggle = gtk_action_group_get_action(win->ag, property);
+    g_object_get(win->conn->gtk_session, property, &state, NULL);
+    gtk_toggle_action_set_active(GTK_TOGGLE_ACTION(toggle), state);
 }
 
 static void menu_cb_toolbar(GtkToggleAction *action, gpointer data)
@@ -546,6 +604,7 @@ static void restore_configuration(struct spice_window *win)
     gchar **keys = NULL;
     gsize nkeys, i;
     GError *error = NULL;
+    gpointer object;
 
     keys = g_key_file_get_keys(keyfile, "general", &nkeys, &error);
     if (error != NULL) {
@@ -567,7 +626,13 @@ static void restore_configuration(struct spice_window *win)
             g_clear_error(&error);
             continue;
         }
-        g_object_set(G_OBJECT(win->spice), keys[i], state, NULL);
+
+        if (is_gtk_session_property(keys[i])) {
+            object = win->conn->gtk_session;
+        } else {
+            object = win->spice;
+        }
+        g_object_set(object, keys[i], state, NULL);
     }
 
     g_strfreev(keys);
@@ -673,6 +738,14 @@ static const GtkActionEntry entries[] = {
     },{
 #endif
 
+#ifdef USE_USBREDIR
+        .name        = "SelectUsbDevices",
+        .label       = N_("_Select USB Devices for redirection"),
+        .callback    = G_CALLBACK(menu_cb_select_usb_devices),
+        .accelerator = "<shift>F10",
+    },{
+#endif
+
         /* Help menu */
         .name        = "About",
         .stock_id    = GTK_STOCK_ABOUT,
@@ -681,12 +754,19 @@ static const GtkActionEntry entries[] = {
     }
 };
 
-static const char *spice_properties[] = {
+static const char *spice_display_properties[] = {
     "grab-keyboard",
     "grab-mouse",
     "resize-guest",
     "scaling",
-    "auto-clipboard"
+    "disable-inputs",
+};
+
+static const char *spice_gtk_session_properties[] = {
+    "auto-clipboard",
+#ifdef USE_USBREDIR
+    "auto-usbredir",
+#endif
 };
 
 static const GtkToggleActionEntry tentries[] = {
@@ -706,11 +786,21 @@ static const GtkToggleActionEntry tentries[] = {
         .name        = "scaling",
         .label       = N_("Scale display"),
         .callback    = G_CALLBACK(menu_cb_bool_prop),
+    },{
+        .name        = "disable-inputs",
+        .label       = N_("Disable inputs"),
+        .callback    = G_CALLBACK(menu_cb_bool_prop),
     },{
         .name        = "auto-clipboard",
         .label       = N_("Automagic clipboard sharing between host and guest"),
         .callback    = G_CALLBACK(menu_cb_bool_prop),
     },{
+#ifdef USE_USBREDIR
+        .name        = "auto-usbredir",
+        .label       = N_("Auto redirect newly plugged in USB devices"),
+        .callback    = G_CALLBACK(menu_cb_bool_prop),
+    },{
+#endif
         .name        = "Statusbar",
         .label       = N_("Statusbar"),
         .callback    = G_CALLBACK(menu_cb_statusbar),
@@ -745,13 +835,20 @@ static char ui_xml[] =
 "      <menuitem action='InsertSmartcard'/>\n"
 "      <menuitem action='RemoveSmartcard'/>\n"
 #endif
+#ifdef USE_USBREDIR
+"      <menuitem action='SelectUsbDevices'/>\n"
+#endif
 "    </menu>\n"
 "    <menu action='OptionMenu'>\n"
 "      <menuitem action='grab-keyboard'/>\n"
 "      <menuitem action='grab-mouse'/>\n"
 "      <menuitem action='resize-guest'/>\n"
 "      <menuitem action='scaling'/>\n"
+"      <menuitem action='disable-inputs'/>\n"
 "      <menuitem action='auto-clipboard'/>\n"
+#ifdef USE_USBREDIR
+"      <menuitem action='auto-usbredir'/>\n"
+#endif
 "    </menu>\n"
 "    <menu action='HelpMenu'>\n"
 "      <menuitem action='About'/>\n"
@@ -767,6 +864,18 @@ static char ui_xml[] =
 "  </toolbar>\n"
 "</ui>\n";
 
+static gboolean is_gtk_session_property(const gchar *property)
+{
+    int i;
+
+    for (i = 0; i < G_N_ELEMENTS(spice_gtk_session_properties); i++) {
+        if (!strcmp(spice_gtk_session_properties[i], property)) {
+            return TRUE;
+        }
+    }
+    return FALSE;
+}
+
 #ifndef WIN32
 static void recent_item_activated_cb(GtkRecentChooser *chooser, gpointer data)
 {
@@ -1011,10 +1120,7 @@ static spice_window *create_spice_window(spice_connection *conn, int id, SpiceCh
     int i;
     SpiceGrabSequence *seq;
 
-    win = malloc(sizeof(*win));
-    if (NULL == win)
-        return NULL;
-    memset(win,0,sizeof(*win));
+    win = g_new0(struct spice_window, 1);
     win->id = id;
     win->conn = conn;
     win->display_channel = channel;
@@ -1116,12 +1222,28 @@ static spice_window *create_spice_window(spice_connection *conn, int id, SpiceCh
     restore_configuration(win);
 
     /* init toggle actions */
-    for (i = 0; i < G_N_ELEMENTS(spice_properties); i++) {
-        toggle = gtk_action_group_get_action(win->ag, spice_properties[i]);
-        g_object_get(win->spice, spice_properties[i], &state, NULL);
+    for (i = 0; i < G_N_ELEMENTS(spice_display_properties); i++) {
+        toggle = gtk_action_group_get_action(win->ag,
+                                             spice_display_properties[i]);
+        g_object_get(win->spice, spice_display_properties[i], &state, NULL);
         gtk_toggle_action_set_active(GTK_TOGGLE_ACTION(toggle), state);
     }
 
+    for (i = 0; i < G_N_ELEMENTS(spice_gtk_session_properties); i++) {
+        char notify[64];
+
+        toggle = gtk_action_group_get_action(win->ag,
+                                             spice_gtk_session_properties[i]);
+        g_object_get(win->conn->gtk_session, spice_gtk_session_properties[i],
+                     &state, NULL);
+        gtk_toggle_action_set_active(GTK_TOGGLE_ACTION(toggle), state);
+
+        snprintf(notify, sizeof(notify), "notify::%s",
+                 spice_gtk_session_properties[i]);
+        g_signal_connect(win->conn->gtk_session, notify,
+                         G_CALLBACK(menu_cb_conn_bool_prop_changed), win);
+    }
+
     toggle = gtk_action_group_get_action(win->ag, "Toolbar");
     state = gtk_widget_get_visible(win->toolbar);
     gtk_toggle_action_set_active(GTK_TOGGLE_ACTION(toggle), state);
@@ -1304,6 +1426,24 @@ static void display_mark(SpiceChannel *channel, gint mark, spice_window *win)
     }
 }
 
+static void update_auto_usbredir_sensitive(spice_connection *conn)
+{
+#ifdef USE_USBREDIR
+    int i;
+    GtkAction *ac;
+    gboolean sensitive;
+
+    sensitive = spice_session_has_channel_type(conn->session,
+                                               SPICE_CHANNEL_USBREDIR);
+    for (i = 0; i < SPICE_N_ELEMENTS(conn->wins); i++) {
+        if (conn->wins[i] == NULL)
+            continue;
+        ac = gtk_action_group_get_action(conn->wins[i]->ag, "auto-usbredir");
+        gtk_action_set_sensitive(ac, sensitive);
+    }
+#endif
+}
+
 static void channel_new(SpiceSession *s, SpiceChannel *channel, gpointer data)
 {
     spice_connection *conn = data;
@@ -1334,6 +1474,7 @@ static void channel_new(SpiceSession *s, SpiceChannel *channel, gpointer data)
         conn->wins[id] = create_spice_window(conn, id, channel);
         g_signal_connect(channel, "display-mark",
                          G_CALLBACK(display_mark), conn->wins[id]);
+        update_auto_usbredir_sensitive(conn);
     }
 
     if (SPICE_IS_INPUTS_CHANNEL(channel)) {
@@ -1348,6 +1489,10 @@ static void channel_new(SpiceSession *s, SpiceChannel *channel, gpointer data)
         SPICE_DEBUG("new audio channel");
         conn->audio = spice_audio_new(s, NULL, NULL);
     }
+
+    if (SPICE_IS_USBREDIR_CHANNEL(channel)) {
+        update_auto_usbredir_sensitive(conn);
+    }
 }
 
 static void channel_destroy(SpiceSession *s, SpiceChannel *channel, gpointer data)
@@ -1378,6 +1523,10 @@ static void channel_destroy(SpiceSession *s, SpiceChannel *channel, gpointer dat
         }
     }
 
+    if (SPICE_IS_USBREDIR_CHANNEL(channel)) {
+        update_auto_usbredir_sensitive(conn);
+    }
+
     conn->channels--;
     if (conn->channels > 0) {
         return;
@@ -1399,15 +1548,24 @@ static void migration_state(GObject *session,
 static spice_connection *connection_new(void)
 {
     spice_connection *conn;
+    SpiceUsbDeviceManager *manager;
 
     conn = g_new0(spice_connection, 1);
     conn->session = spice_session_new();
+    conn->gtk_session = spice_gtk_session_get(conn->session);
     g_signal_connect(conn->session, "channel-new",
                      G_CALLBACK(channel_new), conn);
     g_signal_connect(conn->session, "channel-destroy",
                      G_CALLBACK(channel_destroy), conn);
     g_signal_connect(conn->session, "notify::migration-state",
                      G_CALLBACK(migration_state), conn);
+
+    manager = spice_usb_device_manager_get(conn->session, NULL);
+    if (manager) {
+        g_signal_connect(manager, "auto-connect-failed",
+                         G_CALLBACK(usb_connect_failed), NULL);
+    }
+
     connections++;
     SPICE_DEBUG("%s (%d)", __FUNCTION__, connections);
     return conn;
@@ -1494,6 +1652,25 @@ signal_handler(int signum)
     g_main_loop_quit(mainloop);
 }
 
+static void usb_connect_failed(GObject               *object,
+                               SpiceUsbDevice        *device,
+                               GError                *error,
+                               gpointer               data)
+{
+    GtkWidget *dialog;
+
+    if (error->domain == G_IO_ERROR && error->code == G_IO_ERROR_CANCELLED)
+        return;
+
+    dialog = gtk_message_dialog_new(NULL, GTK_DIALOG_MODAL, GTK_MESSAGE_ERROR,
+                                    GTK_BUTTONS_CLOSE,
+                                    "USB redirection error");
+    gtk_message_dialog_format_secondary_text(GTK_MESSAGE_DIALOG(dialog),
+                                             "%s", error->message);
+    gtk_dialog_run(GTK_DIALOG(dialog));
+    gtk_widget_destroy(dialog);
+}
+
 int main(int argc, char *argv[])
 {
     GError *error = NULL;
