@@ -234,13 +234,11 @@ class EnumBaseType(Type):
         return codegen.prefix_camel(self.name)
 
     def c_enumname(self, value):
-        if self.has_attr("prefix"):
-            return self.attributes["prefix"][0] + self.names[value]
-        return codegen.prefix_underscore_upper(self.name.upper(), self.names[value])
+        self.c_enumname_by_name(self.names[value])
 
     def c_enumname_by_name(self, name):
         if self.has_attr("prefix"):
-            return self.attributes["prefix"][0] + self.names[value]
+            return self.attributes["prefix"][0] + name
         return codegen.prefix_underscore_upper(self.name.upper(), name)
 
     def is_primitive(self):
@@ -909,12 +907,14 @@ class ChannelMember(Containee):
         return "%s (%s)" % (str(self.name), str(self.message_type))
 
 class ChannelType(Type):
-    def __init__(self, name, base, members):
+    def __init__(self, name, base, members, attribute_list):
         Type.__init__(self)
         self.name = name
         self.base = base
         self.member_name = None
         self.members = members
+        for attr in attribute_list:
+            self.attributes[attr[0][1:]] = attr[1:]
 
     def __str__(self):
         if self.name == None:
